"""Message component utilities."""

from __future__ import annotations

from typing import Any, Dict, Optional, TYPE_CHECKING

from .enums import ComponentType, ButtonStyle, ChannelType, TextInputStyle
from .models import (
    ActionRow,
    Button,
    Component,
    SelectMenu,
    SelectOption,
    PartialEmoji,
    PartialEmoji,
    Section,
    TextDisplay,
    Thumbnail,
    MediaGallery,
    MediaGalleryItem,
    FileComponent,
    Separator,
    Container,
    UnfurledMediaItem,
)

if TYPE_CHECKING:  # pragma: no cover - optional client for future use
    from .client import Client


def component_factory(
    data: Dict[str, Any], client: Optional["Client"] = None
) -> "Component":
    """Create a component object from raw API data."""
    ctype = ComponentType(data["type"])

    if ctype == ComponentType.ACTION_ROW:
        row = ActionRow()
        for comp in data.get("components", []):
            row.add_component(component_factory(comp, client))
        return row

    if ctype == ComponentType.BUTTON:
        return Button(
            style=ButtonStyle(data["style"]),
            label=data.get("label"),
            emoji=PartialEmoji(data["emoji"]) if data.get("emoji") else None,
            custom_id=data.get("custom_id"),
            url=data.get("url"),
            disabled=data.get("disabled", False),
        )

    if ctype in {
        ComponentType.STRING_SELECT,
        ComponentType.USER_SELECT,
        ComponentType.ROLE_SELECT,
        ComponentType.MENTIONABLE_SELECT,
        ComponentType.CHANNEL_SELECT,
    }:
        options = [
            SelectOption(
                label=o["label"],
                value=o["value"],
                description=o.get("description"),
                emoji=PartialEmoji(o["emoji"]) if o.get("emoji") else None,
                default=o.get("default", False),
            )
            for o in data.get("options", [])
        ]
        channel_types = None
        if ctype == ComponentType.CHANNEL_SELECT and data.get("channel_types"):
            channel_types = [ChannelType(ct) for ct in data.get("channel_types", [])]

        return SelectMenu(
            custom_id=data["custom_id"],
            options=options,
            placeholder=data.get("placeholder"),
            min_values=data.get("min_values", 1),
            max_values=data.get("max_values", 1),
            disabled=data.get("disabled", False),
            channel_types=channel_types,
            type=ctype,
        )

    if ctype == ComponentType.TEXT_INPUT:
        from .ui.modal import TextInput

        return TextInput(
            label=data.get("label", ""),
            custom_id=data.get("custom_id"),
            style=TextInputStyle(data.get("style", TextInputStyle.SHORT.value)),
            placeholder=data.get("placeholder"),
            required=data.get("required", True),
            min_length=data.get("min_length"),
            max_length=data.get("max_length"),
        )

    if ctype == ComponentType.SECTION:
        # The components in a section can only be TextDisplay
        section_components = []
        for c in data.get("components", []):
            comp = component_factory(c, client)
            if isinstance(comp, TextDisplay):
                section_components.append(comp)

        accessory = None
        if data.get("accessory"):
            acc_comp = component_factory(data["accessory"], client)
            if isinstance(acc_comp, (Thumbnail, Button)):
                accessory = acc_comp

        return Section(
            components=section_components,
            accessory=accessory,
            id=data.get("id"),
        )

    if ctype == ComponentType.TEXT_DISPLAY:
        return TextDisplay(content=data["content"], id=data.get("id"))

    if ctype == ComponentType.THUMBNAIL:
        return Thumbnail(
            media=UnfurledMediaItem(**data["media"]),
            description=data.get("description"),
            spoiler=data.get("spoiler", False),
            id=data.get("id"),
        )

    if ctype == ComponentType.MEDIA_GALLERY:
        return MediaGallery(
            items=[
                MediaGalleryItem(
                    media=UnfurledMediaItem(**i["media"]),
                    description=i.get("description"),
                    spoiler=i.get("spoiler", False),
                )
                for i in data.get("items", [])
            ],
            id=data.get("id"),
        )

    if ctype == ComponentType.FILE:
        return FileComponent(
            file=UnfurledMediaItem(**data["file"]),
            spoiler=data.get("spoiler", False),
            id=data.get("id"),
        )

    if ctype == ComponentType.SEPARATOR:
        return Separator(
            divider=data.get("divider", True),
            spacing=data.get("spacing", 1),
            id=data.get("id"),
        )

    if ctype == ComponentType.CONTAINER:
        return Container(
            components=[
                component_factory(c, client) for c in data.get("components", [])
            ],
            accent_color=data.get("accent_color"),
            spoiler=data.get("spoiler", False),
            id=data.get("id"),
        )

    raise ValueError(f"Unsupported component type: {ctype}")



from __future__ import annotations

from typing import TYPE_CHECKING, Optional, List, Union, Any, Dict

if TYPE_CHECKING:
    from disagreement.client import Client
    from disagreement.interactions import (
        Interaction,
        InteractionCallbackData,
        InteractionResponsePayload,
        Snowflake,
    )
    from disagreement.enums import InteractionCallbackType, MessageFlags
    from disagreement.models import (
        User,
        Member,
        Message,
        Channel,
        ActionRow,
    )
    from disagreement.ui.view import View

    # For full model hints, these would be imported from disagreement.models when defined:
    Embed = Any
    PartialAttachment = Any
    Guild = Any  # from disagreement.models import Guild
    TextChannel = Any  # from disagreement.models import TextChannel, etc.
    from .commands import AppCommand

from disagreement.enums import InteractionCallbackType, MessageFlags
from disagreement.interactions import (
    Interaction,
    InteractionCallbackData,
    InteractionResponsePayload,
    Snowflake,
)
from disagreement.models import Message
from disagreement.typing import Typing


class AppCommandContext:
    """
    Represents the context in which an application command is being invoked.
    Provides methods to respond to the interaction.
    """

    def __init__(
        self,
        bot: "Client",
        interaction: "Interaction",
        command: Optional["AppCommand"] = None,
    ):
        self.bot: "Client" = bot
        self.interaction: "Interaction" = interaction
        self.command: Optional["AppCommand"] = command  # The command that was invoked

        self._responded: bool = False
        self._deferred: bool = False

    @property
    def token(self) -> str:
        """The interaction token."""
        return self.interaction.token

    @property
    def interaction_id(self) -> "Snowflake":
        """The interaction ID."""
        return self.interaction.id

    @property
    def application_id(self) -> "Snowflake":
        """The application ID of the interaction."""
        return self.interaction.application_id

    @property
    def guild_id(self) -> Optional["Snowflake"]:
        """The ID of the guild where the interaction occurred, if any."""
        return self.interaction.guild_id

    @property
    def channel_id(self) -> Optional["Snowflake"]:
        """The ID of the channel where the interaction occurred."""
        return self.interaction.channel_id

    @property
    def author(self) -> Optional[Union["User", "Member"]]:
        """The user or member who invoked the interaction."""
        return self.interaction.member or self.interaction.user

    @property
    def user(self) -> Optional["User"]:
        """The user who invoked the interaction.
        If in a guild, this is the user part of the member.
        If in a DM, this is the top-level user.
        """
        return self.interaction.user

    @property
    def member(self) -> Optional["Member"]:
        """The member who invoked the interaction, if this occurred in a guild."""
        return self.interaction.member

    @property
    def locale(self) -> Optional[str]:
        """The selected language of the invoking user."""
        return self.interaction.locale

    @property
    def guild_locale(self) -> Optional[str]:
        """The guild's preferred language, if applicable."""
        return self.interaction.guild_locale

    @property
    async def guild(self) -> Optional["Guild"]:
        """The guild object where the interaction occurred, if available."""

        if not self.guild_id:
            return None

        guild = None
        if hasattr(self.bot, "get_guild"):
            guild = self.bot.get_guild(self.guild_id)

        if not guild and hasattr(self.bot, "fetch_guild"):
            try:
                guild = await self.bot.fetch_guild(self.guild_id)
            except Exception:
                guild = None

        return guild

    @property
    async def channel(self) -> Optional[Any]:
        """The channel object where the interaction occurred, if available."""

        if not self.channel_id:
            return None

        channel = None
        if hasattr(self.bot, "get_channel"):
            channel = self.bot.get_channel(self.channel_id)
        elif hasattr(self.bot, "_channels"):
            channel = self.bot._channels.get(self.channel_id)

        if not channel and hasattr(self.bot, "fetch_channel"):
            try:
                channel = await self.bot.fetch_channel(self.channel_id)
            except Exception:
                channel = None

        return channel

    async def _send_response(
        self,
        response_type: "InteractionCallbackType",
        data: Optional[Dict[str, Any]] = None,
    ) -> None:
        """Internal helper to send interaction responses."""
        if (
            self._responded
            and not self._deferred
            and response_type
            != InteractionCallbackType.APPLICATION_COMMAND_AUTOCOMPLETE_RESULT
        ):
            # If already responded and not deferred, subsequent responses must be followups
            # (unless it's an autocomplete result which is a special case)
            # For now, let's assume followups are handled by separate methods.
            # This logic might need refinement based on how followups are exposed.
            raise RuntimeError(
                "Interaction has already been responded to. Use send_followup()."
            )

        callback_data = InteractionCallbackData(data) if data else None
        payload = InteractionResponsePayload(type=response_type, data=callback_data)

        await self.bot._http.create_interaction_response(
            interaction_id=self.interaction_id,
            interaction_token=self.token,
            payload=payload,
        )
        if (
            response_type
            != InteractionCallbackType.APPLICATION_COMMAND_AUTOCOMPLETE_RESULT
        ):
            self._responded = True
            if (
                response_type
                == InteractionCallbackType.DEFERRED_CHANNEL_MESSAGE_WITH_SOURCE
                or response_type == InteractionCallbackType.DEFERRED_UPDATE_MESSAGE
            ):
                self._deferred = True

    async def defer(self, ephemeral: bool = False, thinking: bool = True) -> None:
        """
        Defers the interaction response.

        This is typically used when your command might take longer than 3 seconds to process.
        You must send a followup message within 15 minutes.

        Args:
            ephemeral (bool): Whether the subsequent followup response should be ephemeral.
                              Only applicable if `thinking` is True.
            thinking (bool): If True (default), responds with a "Bot is thinking..." message
                             (DEFERRED_CHANNEL_MESSAGE_WITH_SOURCE).
                             If False, responds with DEFERRED_UPDATE_MESSAGE (for components).
        """
        if self._responded:
            raise RuntimeError("Interaction has already been responded to or deferred.")

        response_type = (
            InteractionCallbackType.DEFERRED_CHANNEL_MESSAGE_WITH_SOURCE
            if thinking
            else InteractionCallbackType.DEFERRED_UPDATE_MESSAGE
        )
        data = None
        if ephemeral and thinking:
            data = {
                "flags": MessageFlags.EPHEMERAL.value
            }  # Assuming MessageFlags enum exists

        await self._send_response(response_type, data)
        self._deferred = True  # Mark as deferred

    async def send(
        self,
        content: Optional[str] = None,
        embed: Optional["Embed"] = None,  # Convenience for single embed
        embeds: Optional[List["Embed"]] = None,
        *,
        tts: bool = False,
        files: Optional[List[Any]] = None,
        components: Optional[List[ActionRow]] = None,
        view: Optional[View] = None,
        allowed_mentions: Optional[Dict[str, Any]] = None,
        ephemeral: bool = False,
        flags: Optional[int] = None,
    ) -> Optional[
        "Message"
    ]:  # Returns Message if not ephemeral and response was not deferred
        """
        Sends a response to the interaction.
        If the interaction was previously deferred, this will edit the original deferred response.
        Otherwise, it sends an initial response.

        Args:
            content (Optional[str]): The message content.
            embed (Optional[Embed]): A single embed to send. If `embeds` is also provided, this is ignored.
            embeds (Optional[List[Embed]]): A list of embeds to send (max 10).
            ephemeral (bool): Whether the message should be ephemeral (only visible to the invoker).
            flags (Optional[int]): Additional message flags to apply.

        Returns:
            Optional[Message]: The sent message object if a new message was created and not ephemeral.
                               None if the response was ephemeral or an edit to a deferred message.
        """
        if not self._responded and self._deferred:  # Editing a deferred response
            # Use edit_original_interaction_response
            payload: Dict[str, Any] = {}
            if content is not None:
                payload["content"] = content

            if tts:
                payload["tts"] = True

            actual_embeds = embeds
            if embed and not embeds:
                actual_embeds = [embed]
            if actual_embeds:
                payload["embeds"] = [e.to_dict() for e in actual_embeds]

            if view:
                await view._start(self.bot)
                payload["components"] = view.to_components_payload()
            elif components:
                payload["components"] = [c.to_dict() for c in components]

            if files is not None:
                payload["attachments"] = [
                    f.to_dict() if hasattr(f, "to_dict") else f for f in files
                ]

            if allowed_mentions is not None:
                payload["allowed_mentions"] = allowed_mentions

            # Flags (like ephemeral) cannot be set when editing the original deferred response this way.
            # Ephemeral for deferred must be set during defer().

            msg_data = await self.bot._http.edit_original_interaction_response(
                application_id=self.application_id,
                interaction_token=self.token,
                payload=payload,
            )
            self._responded = True  # Ensure it's marked as fully responded
            if view and msg_data and "id" in msg_data:
                view.message_id = msg_data["id"]
                self.bot._views[msg_data["id"]] = view
            # Construct and return Message object if needed, for now returns None for edits
            return None

        elif not self._responded:  # Sending an initial response
            data: Dict[str, Any] = {}
            if content is not None:
                data["content"] = content

            if tts:
                data["tts"] = True

            actual_embeds = embeds
            if embed and not embeds:
                actual_embeds = [embed]
            if actual_embeds:
                data["embeds"] = [
                    e.to_dict() for e in actual_embeds
                ]  # Assuming embeds have to_dict()

            if view:
                await view._start(self.bot)
                data["components"] = view.to_components_payload()
            elif components:
                data["components"] = [c.to_dict() for c in components]

            if files is not None:
                data["attachments"] = [
                    f.to_dict() if hasattr(f, "to_dict") else f for f in files
                ]

            if allowed_mentions is not None:
                data["allowed_mentions"] = allowed_mentions

            flags_value = 0
            if ephemeral:
                flags_value |= MessageFlags.EPHEMERAL.value
            if flags:
                flags_value |= flags
            if flags_value:
                data["flags"] = flags_value

            await self._send_response(
                InteractionCallbackType.CHANNEL_MESSAGE_WITH_SOURCE, data
            )

            if view and not ephemeral:
                try:
                    msg_data = await self.bot._http.get_original_interaction_response(
                        application_id=self.application_id,
                        interaction_token=self.token,
                    )
                    if msg_data and "id" in msg_data:
                        view.message_id = msg_data["id"]
                        self.bot._views[msg_data["id"]] = view
                except Exception:
                    pass
            if not ephemeral:
                return None
            return None
        else:
            # If already responded and not deferred, this should be a followup.
            # This method is for initial response or editing deferred.
            raise RuntimeError(
                "Interaction has already been responded to. Use send_followup()."
            )

    async def send_followup(
        self,
        content: Optional[str] = None,
        embed: Optional["Embed"] = None,
        embeds: Optional[List["Embed"]] = None,
        *,
        ephemeral: bool = False,
        tts: bool = False,
        files: Optional[List[Any]] = None,
        components: Optional[List["ActionRow"]] = None,
        view: Optional[View] = None,
        allowed_mentions: Optional[Dict[str, Any]] = None,
        flags: Optional[int] = None,
    ) -> Optional["Message"]:
        """
        Sends a followup message to an interaction.
        This can be used after an initial response or a deferred response.

        Args:
            content (Optional[str]): The message content.
            embed (Optional[Embed]): A single embed to send.
            embeds (Optional[List[Embed]]): A list of embeds to send.
            ephemeral (bool): Whether the followup message should be ephemeral.
            flags (Optional[int]): Additional message flags to apply.

        Returns:
            Message: The sent followup message object.
        """
        if not self._responded:
            raise RuntimeError(
                "Must acknowledge or defer the interaction before sending a followup."
            )

        payload: Dict[str, Any] = {}
        if content is not None:
            payload["content"] = content

        if tts:
            payload["tts"] = True

        actual_embeds = embeds
        if embed and not embeds:
            actual_embeds = [embed]
        if actual_embeds:
            payload["embeds"] = [
                e.to_dict() for e in actual_embeds
            ]  # Assuming embeds have to_dict()

        if view:
            await view._start(self.bot)
            payload["components"] = view.to_components_payload()
        elif components:
            payload["components"] = [c.to_dict() for c in components]

        if files is not None:
            payload["attachments"] = [
                f.to_dict() if hasattr(f, "to_dict") else f for f in files
            ]

        if allowed_mentions is not None:
            payload["allowed_mentions"] = allowed_mentions

        flags_value = 0
        if ephemeral:
            flags_value |= MessageFlags.EPHEMERAL.value
        if flags:
            flags_value |= flags
        if flags_value:
            payload["flags"] = flags_value

        # Followup messages are sent to a webhook endpoint
        message_data = await self.bot._http.create_followup_message(
            application_id=self.application_id,
            interaction_token=self.token,
            payload=payload,
        )
        if view and message_data and "id" in message_data:
            view.message_id = message_data["id"]
            self.bot._views[message_data["id"]] = view
        from disagreement.models import Message  # Ensure Message is available

        return Message(data=message_data, client_instance=self.bot)

    async def edit(
        self,
        message_id: "Snowflake" = "@original",  # Defaults to editing the original response
        content: Optional[str] = None,
        embed: Optional["Embed"] = None,
        embeds: Optional[List["Embed"]] = None,
        *,
        components: Optional[List["ActionRow"]] = None,
        attachments: Optional[List[Any]] = None,
        allowed_mentions: Optional[Dict[str, Any]] = None,
    ) -> Optional["Message"]:
        """
        Edits a message previously sent in response to this interaction.
        Can edit the original response or a followup message.

        Args:
            message_id (Snowflake): The ID of the message to edit. Defaults to "@original"
                                    to edit the initial interaction response.
            content (Optional[str]): The new message content.
            embed (Optional[Embed]): A single new embed.
            embeds (Optional[List[Embed]]): A list of new embeds.

        Returns:
            Optional[Message]: The edited message object if available.
        """
        if not self._responded:
            raise RuntimeError(
                "Cannot edit response if interaction hasn't been responded to or deferred."
            )

        payload: Dict[str, Any] = {}
        if content is not None:
            payload["content"] = content  # Use None to clear

        actual_embeds = embeds
        if embed and not embeds:
            actual_embeds = [embed]
        if actual_embeds is not None:  # Allow passing empty list to clear embeds
            payload["embeds"] = [
                e.to_dict() for e in actual_embeds
            ]  # Assuming embeds have to_dict()

        if components is not None:
            payload["components"] = [c.to_dict() for c in components]

        if attachments is not None:
            payload["attachments"] = [
                a.to_dict() if hasattr(a, "to_dict") else a for a in attachments
            ]

        if allowed_mentions is not None:
            payload["allowed_mentions"] = allowed_mentions

        if message_id == "@original":
            edited_message_data = (
                await self.bot._http.edit_original_interaction_response(
                    application_id=self.application_id,
                    interaction_token=self.token,
                    payload=payload,
                )
            )
        else:
            edited_message_data = await self.bot._http.edit_followup_message(
                application_id=self.application_id,
                interaction_token=self.token,
                message_id=message_id,
                payload=payload,
            )
        # The HTTP methods used in tests return minimal data that is insufficient
        # to construct a full ``Message`` instance, so we simply return ``None``
        # rather than attempting to parse the response.
        return None

    async def delete(self, message_id: "Snowflake" = "@original") -> None:
        """
        Deletes a message previously sent in response to this interaction.
        Can delete the original response or a followup message.

        Args:
            message_id (Snowflake): The ID of the message to delete. Defaults to "@original"
                                    to delete the initial interaction response.
        """
        if not self._responded:
            # If not responded, there's nothing to delete via this interaction's lifecycle.
            # Deferral doesn't create a message to delete until a followup is sent.
            raise RuntimeError(
                "Cannot delete response if interaction hasn't been responded to."
            )

        if message_id == "@original":
            await self.bot._http.delete_original_interaction_response(
                application_id=self.application_id, interaction_token=self.token
            )
        else:
            await self.bot._http.delete_followup_message(
                application_id=self.application_id,
                interaction_token=self.token,
                message_id=message_id,
            )
        # After deleting the original response, further followups might be problematic.
        # Discord docs: "Once the original message is deleted, you can no longer edit the message or send followups."
        # Consider implications for context state.

    def typing(self) -> Typing:
        """Return a typing context manager for this interaction's channel."""

        if not self.channel_id:
            raise RuntimeError("Cannot send typing indicator without a channel.")
        return self.bot.typing(self.channel_id)



"""
Custom exceptions for the command extension.
"""

from disagreement.errors import DisagreementException


class CommandError(DisagreementException):
    """Base exception for errors raised by the commands extension."""

    pass


class CommandNotFound(CommandError):
    """Exception raised when a command is not found."""

    def __init__(self, command_name: str):
        self.command_name = command_name
        super().__init__(f"Command '{command_name}' not found.")


class BadArgument(CommandError):
    """Exception raised when a command argument fails to parse or validate."""

    pass


class MissingRequiredArgument(BadArgument):
    """Exception raised when a required command argument is missing."""

    def __init__(self, param_name: str):
        self.param_name = param_name
        super().__init__(f"Missing required argument: {param_name}")


class ArgumentParsingError(BadArgument):
    """Exception raised during the argument parsing process."""

    pass


class CheckFailure(CommandError):
    """Exception raised when a command check fails."""

    pass


class CheckAnyFailure(CheckFailure):
    """Raised when :func:`check_any` fails all checks."""

    def __init__(self, errors: list[CheckFailure]):
        self.errors = errors
        msg = "; ".join(str(e) for e in errors)
        super().__init__(f"All checks failed: {msg}")


class CommandOnCooldown(CheckFailure):
    """Raised when a command is invoked while on cooldown."""

    def __init__(self, retry_after: float):
        self.retry_after = retry_after
        super().__init__(f"Command is on cooldown. Retry in {retry_after:.2f}s")


class CommandInvokeError(CommandError):
    """Exception raised when an error occurs during command invocation."""

    def __init__(self, original: Exception):
        self.original = original
        super().__init__(f"Error during command invocation: {original}")


class MaxConcurrencyReached(CommandError):
    """Raised when a command exceeds its concurrency limit."""

    def __init__(self, limit: int):
        self.limit = limit
        super().__init__(f"Max concurrency of {limit} reached")


# Add more specific errors as needed, e.g., UserNotFound, ChannelNotFound, etc.
# These might inherit from BadArgument.



from __future__ import annotations

import time
from typing import TYPE_CHECKING, Callable, Dict, Generic, Optional, TypeVar
from collections import OrderedDict

if TYPE_CHECKING:
    from .models import Channel, Guild, Member
    from .caching import MemberCacheFlags

T = TypeVar("T")


class Cache(Generic[T]):
    """Simple in-memory cache with optional TTL and max size support."""

    def __init__(
        self, ttl: Optional[float] = None, maxlen: Optional[int] = None
    ) -> None:
        self.ttl = ttl
        self.maxlen = maxlen
        self._data: "OrderedDict[str, tuple[T, Optional[float]]]" = OrderedDict()

    def set(self, key: str, value: T) -> None:
        expiry = time.monotonic() + self.ttl if self.ttl is not None else None
        if key in self._data:
            self._data.move_to_end(key)
        self._data[key] = (value, expiry)
        if self.maxlen is not None and len(self._data) > self.maxlen:
            self._data.popitem(last=False)

    def get(self, key: str) -> Optional[T]:
        item = self._data.get(key)
        if not item:
            return None
        value, expiry = item
        if expiry is not None and expiry < time.monotonic():
            self.invalidate(key)
            return None
        self._data.move_to_end(key)
        return value

    def get_or_fetch(self, key: str, fetch_fn: Callable[[], T]) -> T:
        """Return a cached item or fetch and store it if missing."""
        value = self.get(key)
        if value is None:
            value = fetch_fn()
            self.set(key, value)
        return value

    def invalidate(self, key: str) -> None:
        self._data.pop(key, None)

    def clear(self) -> None:
        self._data.clear()

    def values(self) -> list[T]:
        now = time.monotonic()
        items = []
        for key, (value, expiry) in list(self._data.items()):
            if expiry is not None and expiry < now:
                self.invalidate(key)
            else:
                items.append(value)
        return items


class GuildCache(Cache["Guild"]):
    """Cache specifically for :class:`Guild` objects."""


class ChannelCache(Cache["Channel"]):
    """Cache specifically for :class:`Channel` objects."""


class MemberCache(Cache["Member"]):
    """
    A cache for :class:`Member` objects that respects :class:`MemberCacheFlags`.
    """

    def __init__(self, flags: MemberCacheFlags, ttl: Optional[float] = None) -> None:
        super().__init__(ttl)
        self.flags = flags

    def _should_cache(self, member: Member) -> bool:
        """Determines if a member should be cached based on the flags."""
        if self.flags.all:
            return True
        if self.flags.none:
            return False

        if self.flags.online and member.status != "offline":
            return True
        if self.flags.voice and member.voice_state is not None:
            return True
        if self.flags.joined and getattr(member, "_just_joined", False):
            return True
        return False

    def set(self, key: str, value: Member) -> None:
        if self._should_cache(value):
            super().set(key, value)



import pytest

from disagreement.interactions import Interaction
from disagreement.ui import Modal, text_input
from disagreement.enums import InteractionCallbackType, TextInputStyle


class MyModal(Modal):
    def __init__(self):
        super().__init__(title="Test", custom_id="m1")

    @text_input(label="Name", style=TextInputStyle.SHORT)
    async def name(self, interaction: Interaction):
        pass


def test_modal_to_dict():
    modal = MyModal()
    data = modal.to_dict()
    assert data["title"] == "Test"
    assert data["custom_id"] == "m1"
    assert data["components"][0]["components"][0]["label"] == "Name"


@pytest.mark.asyncio
async def test_respond_modal(dummy_bot, interaction):
    modal = MyModal()
    await interaction.respond_modal(modal)
    dummy_bot._http.create_interaction_response.assert_called_once()
    payload = dummy_bot._http.create_interaction_response.call_args.kwargs["payload"]
    assert payload.type == InteractionCallbackType.MODAL
    assert payload.data["custom_id"] == "m1"



import pytest
from types import SimpleNamespace
from unittest.mock import AsyncMock

from disagreement.http import HTTPClient
from disagreement.client import Client


@pytest.mark.asyncio
async def test_get_guild_widget_calls_request():
    http = HTTPClient(token="t")
    http.request = AsyncMock(return_value={})
    await http.get_guild_widget("1")
    http.request.assert_called_once_with("GET", "/guilds/1/widget")


@pytest.mark.asyncio
async def test_edit_guild_widget_calls_request():
    http = HTTPClient(token="t")
    http.request = AsyncMock(return_value={})
    payload = {"enabled": True}
    await http.edit_guild_widget("1", payload)
    http.request.assert_called_once_with("PATCH", "/guilds/1/widget", payload=payload)


@pytest.mark.asyncio
async def test_client_fetch_widget_returns_data():
    http = SimpleNamespace(get_guild_widget=AsyncMock(return_value={"enabled": True}))
    client = Client.__new__(Client)
    client._http = http
    client._closed = False

    data = await client.fetch_widget("1")

    http.get_guild_widget.assert_awaited_once_with("1")
    assert data == {"enabled": True}


@pytest.mark.asyncio
async def test_client_edit_widget_returns_data():
    http = SimpleNamespace(edit_guild_widget=AsyncMock(return_value={"enabled": False}))
    client = Client.__new__(Client)
    client._http = http
    client._closed = False

    payload = {"enabled": False}
    data = await client.edit_widget("1", payload)

    http.edit_guild_widget.assert_awaited_once_with("1", payload)
    assert data == {"enabled": False}



import asyncio
import pytest

from disagreement.ext.commands.core import Command, CommandContext
from disagreement.ext.commands.decorators import (
    check,
    cooldown,
    requires_permissions,
)
from disagreement.ext.commands.errors import CheckFailure, CommandOnCooldown
from disagreement.permissions import Permissions


@pytest.mark.asyncio
async def test_check_decorator_blocks(message):
    async def cb(ctx):
        pass

    cmd = Command(check(lambda c: False)(cb))
    ctx = CommandContext(
        message=message,
        bot=message._client,
        prefix="!",
        command=cmd,
        invoked_with="test",
    )

    with pytest.raises(CheckFailure):
        await cmd.invoke(ctx)


@pytest.mark.asyncio
async def test_cooldown_per_user(message):
    uses = []

    @cooldown(1, 0.1)
    async def cb(ctx):
        uses.append(1)

    cmd = Command(cb)
    ctx = CommandContext(
        message=message,
        bot=message._client,
        prefix="!",
        command=cmd,
        invoked_with="test",
    )

    await cmd.invoke(ctx)

    with pytest.raises(CommandOnCooldown):
        await cmd.invoke(ctx)

    await asyncio.sleep(0.1)
    await cmd.invoke(ctx)
    assert len(uses) == 2


@pytest.mark.asyncio
@pytest.mark.asyncio
async def test_requires_permissions_pass(message):
    class Guild:
        id = "g"
        owner_id = "owner"
        roles = []

        def get_member(self, mid):
            return message.author

    class Channel:
        def __init__(self, perms):
            self.perms = perms
            self.guild_id = "g"

        def permissions_for(self, member):
            return self.perms

    message.author.roles = []
    message._client.get_channel = lambda cid: Channel(Permissions.SEND_MESSAGES)
    message._client.get_guild = lambda gid: Guild()

    @requires_permissions(Permissions.SEND_MESSAGES)
    async def cb(ctx):
        pass

    cmd = Command(cb)
    ctx = CommandContext(
        message=message,
        bot=message._client,
        prefix="!",
        command=cmd,
        invoked_with="test",
    )

    await cmd.invoke(ctx)


@pytest.mark.asyncio
@pytest.mark.asyncio
async def test_requires_permissions_fail(message):
    class Guild:
        id = "g"
        owner_id = "owner"
        roles = []

        def get_member(self, mid):
            return message.author

    class Channel:
        def __init__(self, perms):
            self.perms = perms
            self.guild_id = "g"

        def permissions_for(self, member):
            return self.perms

    message.author.roles = []
    message._client.get_channel = lambda cid: Channel(Permissions.SEND_MESSAGES)
    message._client.get_guild = lambda gid: Guild()

    @requires_permissions(Permissions.MANAGE_MESSAGES)
    async def cb(ctx):
        pass

    cmd = Command(cb)
    ctx = CommandContext(
        message=message,
        bot=message._client,
        prefix="!",
        command=cmd,
        invoked_with="test",
    )

    with pytest.raises(CheckFailure):
        await cmd.invoke(ctx)



import asyncio
from contextlib import suppress
from typing import Optional, TYPE_CHECKING

from .errors import DisagreementException

if TYPE_CHECKING:
    from .client import Client

if __name__ == "typing":
    # For direct module execution testing
    pass


class Typing:
    """Async context manager for Discord typing indicator."""

    def __init__(self, client: "Client", channel_id: str) -> None:
        self._client = client
        self._channel_id = channel_id
        self._task: Optional[asyncio.Task] = None

    async def _run(self) -> None:
        try:
            while True:
                await self._client._http.trigger_typing(self._channel_id)
                await asyncio.sleep(5)
        except asyncio.CancelledError:
            pass

    async def __aenter__(self) -> "Typing":
        if self._client._closed:
            raise DisagreementException("Client is closed.")
        await self._client._http.trigger_typing(self._channel_id)
        self._task = asyncio.create_task(self._run())
        return self

    async def __aexit__(self, exc_type, exc, tb) -> None:
        if self._task:
            self._task.cancel()
            with suppress(asyncio.CancelledError):
                await self._task



"""
Data models for Discord Interaction objects.
"""

from typing import Optional, List, Dict, Union, Any, TYPE_CHECKING

from .enums import (
    ApplicationCommandType,
    ApplicationCommandOptionType,
    InteractionType,
    InteractionCallbackType,
    IntegrationType,
    InteractionContextType,
    ChannelType,
)

# Runtime imports for models used in this module
from .models import (
    User,
    Message,
    Member,
    Role,
    Embed,
    PartialChannel,
    Attachment,
    ActionRow,
    Component,
    AllowedMentions,
)

if TYPE_CHECKING:
    # Import Client type only for type checking to avoid circular imports
    from .client import Client
    from .ui.modal import Modal

    # MessageFlags, PartialAttachment can be added if/when defined

Snowflake = str


# Based on Application Command Option Choice Structure
class ApplicationCommandOptionChoice:
    """Represents a choice for an application command option."""

    def __init__(self, data: dict):
        self.name: str = data["name"]
        self.value: Union[str, int, float] = data["value"]
        self.name_localizations: Optional[Dict[str, str]] = data.get(
            "name_localizations"
        )

    def __repr__(self) -> str:
        return (
            f"<ApplicationCommandOptionChoice name='{self.name}' value={self.value!r}>"
        )

    def to_dict(self) -> Dict[str, Any]:
        payload: Dict[str, Any] = {"name": self.name, "value": self.value}
        if self.name_localizations:
            payload["name_localizations"] = self.name_localizations
        return payload


# Based on Application Command Option Structure
class ApplicationCommandOption:
    """Represents an option for an application command."""

    def __init__(self, data: dict):
        self.type: ApplicationCommandOptionType = ApplicationCommandOptionType(
            data["type"]
        )
        self.name: str = data["name"]
        self.description: str = data["description"]
        self.required: bool = data.get("required", False)

        self.choices: Optional[List[ApplicationCommandOptionChoice]] = (
            [ApplicationCommandOptionChoice(c) for c in data["choices"]]
            if data.get("choices")
            else None
        )

        self.options: Optional[List["ApplicationCommandOption"]] = (
            [ApplicationCommandOption(o) for o in data["options"]]
            if data.get("options")
            else None
        )  # For subcommands/groups

        self.channel_types: Optional[List[ChannelType]] = (
            [ChannelType(ct) for ct in data.get("channel_types", [])]
            if data.get("channel_types")
            else None
        )
        self.min_value: Optional[Union[int, float]] = data.get("min_value")
        self.max_value: Optional[Union[int, float]] = data.get("max_value")
        self.min_length: Optional[int] = data.get("min_length")
        self.max_length: Optional[int] = data.get("max_length")
        self.autocomplete: bool = data.get("autocomplete", False)
        self.name_localizations: Optional[Dict[str, str]] = data.get(
            "name_localizations"
        )
        self.description_localizations: Optional[Dict[str, str]] = data.get(
            "description_localizations"
        )

    def __repr__(self) -> str:
        return f"<ApplicationCommandOption name='{self.name}' type={self.type!r}>"

    def to_dict(self) -> Dict[str, Any]:
        payload: Dict[str, Any] = {
            "type": self.type.value,
            "name": self.name,
            "description": self.description,
        }
        if self.required:  # Defaults to False, only include if True
            payload["required"] = self.required
        if self.choices:
            payload["choices"] = [c.to_dict() for c in self.choices]
        if self.options:  # For subcommands/groups
            payload["options"] = [o.to_dict() for o in self.options]
        if self.channel_types:
            payload["channel_types"] = [ct.value for ct in self.channel_types]
        if self.min_value is not None:
            payload["min_value"] = self.min_value
        if self.max_value is not None:
            payload["max_value"] = self.max_value
        if self.min_length is not None:
            payload["min_length"] = self.min_length
        if self.max_length is not None:
            payload["max_length"] = self.max_length
        if self.autocomplete:  # Defaults to False, only include if True
            payload["autocomplete"] = self.autocomplete
        if self.name_localizations:
            payload["name_localizations"] = self.name_localizations
        if self.description_localizations:
            payload["description_localizations"] = self.description_localizations
        return payload


# Based on Application Command Structure
class ApplicationCommand:
    """Represents an application command."""

    def __init__(self, data: dict):
        self.id: Optional[Snowflake] = data.get("id")
        self.type: ApplicationCommandType = ApplicationCommandType(
            data.get("type", 1)
        )  # Default to CHAT_INPUT
        self.application_id: Optional[Snowflake] = data.get("application_id")
        self.guild_id: Optional[Snowflake] = data.get("guild_id")
        self.name: str = data["name"]
        self.description: str = data.get(
            "description", ""
        )  # Empty for USER/MESSAGE commands

        self.options: Optional[List[ApplicationCommandOption]] = (
            [ApplicationCommandOption(o) for o in data["options"]]
            if data.get("options")
            else None
        )

        self.default_member_permissions: Optional[str] = data.get(
            "default_member_permissions"
        )
        self.dm_permission: Optional[bool] = data.get("dm_permission")  # Deprecated
        self.nsfw: bool = data.get("nsfw", False)
        self.version: Optional[Snowflake] = data.get("version")
        self.name_localizations: Optional[Dict[str, str]] = data.get(
            "name_localizations"
        )
        self.description_localizations: Optional[Dict[str, str]] = data.get(
            "description_localizations"
        )

        self.integration_types: Optional[List[IntegrationType]] = (
            [IntegrationType(it) for it in data["integration_types"]]
            if data.get("integration_types")
            else None
        )

        self.contexts: Optional[List[InteractionContextType]] = (
            [InteractionContextType(c) for c in data["contexts"]]
            if data.get("contexts")
            else None
        )

    def __repr__(self) -> str:
        return (
            f"<ApplicationCommand id='{self.id}' name='{self.name}' type={self.type!r}>"
        )


# Based on Interaction Object's Resolved Data Structure
class ResolvedData:
    """Represents resolved data for an interaction."""

    def __init__(
        self, data: dict, client_instance: Optional["Client"] = None
    ):  # client_instance for model hydration
        # Models are now imported in TYPE_CHECKING block

        users_data = data.get("users", {})
        self.users: Dict[Snowflake, "User"] = {
            uid: User(udata) for uid, udata in users_data.items()
        }

        self.members: Dict[Snowflake, "Member"] = {}
        for mid, mdata in data.get("members", {}).items():
            member_payload = dict(mdata)
            member_payload.setdefault("id", mid)
            if "user" not in member_payload and mid in users_data:
                member_payload["user"] = users_data[mid]
            self.members[mid] = Member(member_payload, client_instance=client_instance)

        self.roles: Dict[Snowflake, "Role"] = {
            rid: Role(rdata) for rid, rdata in data.get("roles", {}).items()
        }

        self.channels: Dict[Snowflake, "PartialChannel"] = {
            cid: PartialChannel(cdata, client_instance=client_instance)
            for cid, cdata in data.get("channels", {}).items()
        }

        self.messages: Dict[Snowflake, "Message"] = (
            {
                mid: Message(mdata, client_instance=client_instance) for mid, mdata in data.get("messages", {}).items()  # type: ignore[misc]
            }
            if client_instance
            else {}
        )  # Only hydrate if client is available

        self.attachments: Dict[Snowflake, "Attachment"] = {
            aid: Attachment(adata) for aid, adata in data.get("attachments", {}).items()
        }

    def __repr__(self) -> str:
        return f"<ResolvedData users={len(self.users)} members={len(self.members)} roles={len(self.roles)} channels={len(self.channels)} messages={len(self.messages)} attachments={len(self.attachments)}>"


# Based on Interaction Object's Data Structure (for Application Commands)
class InteractionData:
    """Represents the data payload for an interaction."""

    def __init__(self, data: dict, client_instance: Optional["Client"] = None):
        self.id: Optional[Snowflake] = data.get("id")  # Command ID
        self.name: Optional[str] = data.get("name")  # Command name
        self.type: Optional[ApplicationCommandType] = (
            ApplicationCommandType(data["type"]) if data.get("type") else None
        )

        self.resolved: Optional[ResolvedData] = (
            ResolvedData(data["resolved"], client_instance=client_instance)
            if data.get("resolved")
            else None
        )

        # For CHAT_INPUT, this is List[ApplicationCommandInteractionDataOption]
        # For USER/MESSAGE, this is not present or different.
        # For now, storing as raw list of dicts. Parsing can happen in handler.
        self.options: Optional[List[Dict[str, Any]]] = data.get("options")

        # For message components
        self.custom_id: Optional[str] = data.get("custom_id")
        self.component_type: Optional[int] = data.get("component_type")
        self.values: Optional[List[str]] = data.get("values")

        self.guild_id: Optional[Snowflake] = data.get("guild_id")
        self.target_id: Optional[Snowflake] = data.get(
            "target_id"
        )  # For USER/MESSAGE commands

    def __repr__(self) -> str:
        return f"<InteractionData id='{self.id}' name='{self.name}' type={self.type!r}>"


# Based on Interaction Object Structure
class Interaction:
    """Represents an interaction from Discord."""

    def __init__(self, data: dict, client_instance: "Client"):
        self._client: "Client" = client_instance

        self.id: Snowflake = data["id"]
        self.application_id: Snowflake = data["application_id"]
        self.type: InteractionType = InteractionType(data["type"])

        self.data: Optional[InteractionData] = (
            InteractionData(data["data"], client_instance=client_instance)
            if data.get("data")
            else None
        )

        self.guild_id: Optional[Snowflake] = data.get("guild_id")
        self.channel_id: Optional[Snowflake] = data.get(
            "channel_id"
        )  # Will be present on command invocations

        member_data = data.get("member")
        user_data_from_member = (
            member_data.get("user") if isinstance(member_data, dict) else None
        )

        self.member: Optional["Member"] = (
            Member(member_data, client_instance=self._client) if member_data else None
        )

        # User object is included within member if in guild, otherwise it's top-level
        # If self.member was successfully hydrated, its .user attribute should be preferred if it exists.
        # However, Member.__init__ handles setting User attributes.
        # The primary source for User is data.get("user") or member_data.get("user").

        if data.get("user"):
            self.user: Optional["User"] = User(data["user"])
        elif user_data_from_member:
            self.user: Optional["User"] = User(user_data_from_member)
        elif (
            self.member
        ):  # If member was hydrated and has user attributes (e.g. from Member(User) inheritance)
            # This assumes Member correctly populates its User parts.
            self.user: Optional["User"] = self.member  # Member is a User subclass
        else:
            self.user: Optional["User"] = None

        self.token: str = data["token"]  # For responding to the interaction
        self.version: int = data["version"]

        self.message: Optional["Message"] = (
            Message(data["message"], client_instance=client_instance)
            if data.get("message")
            else None
        )  # For component interactions

        self.app_permissions: Optional[str] = data.get(
            "app_permissions"
        )  # Bitwise set of permissions the app has in the source channel
        self.locale: Optional[str] = data.get(
            "locale"
        )  # Selected language of the invoking user
        self.guild_locale: Optional[str] = data.get(
            "guild_locale"
        )  # Guild's preferred language

        self.response = InteractionResponse(self)

    async def respond(
        self,
        content: Optional[str] = None,
        *,
        embed: Optional[Embed] = None,
        embeds: Optional[List[Embed]] = None,
        components: Optional[List[ActionRow]] = None,
        ephemeral: bool = False,
        tts: bool = False,
    ) -> None:
        """|coro|

        Responds to this interaction.

        Parameters:
            content (Optional[str]): The content of the message.
            embed (Optional[Embed]): A single embed to send.
            embeds (Optional[List[Embed]]): A list of embeds to send.
            components (Optional[List[ActionRow]]): A list of ActionRow components.
            ephemeral (bool): Whether the response should be ephemeral (only visible to the user).
            tts (bool): Whether the message should be sent with text-to-speech.
        """
        if embed and embeds:
            raise ValueError("Cannot provide both embed and embeds.")

        data: Dict[str, Any] = {}
        if tts:
            data["tts"] = True
        if content:
            data["content"] = content
        if embed:
            data["embeds"] = [embed.to_dict()]
        elif embeds:
            data["embeds"] = [e.to_dict() for e in embeds]
        if components:
            data["components"] = [c.to_dict() for c in components]
        if ephemeral:
            data["flags"] = 1 << 6  # EPHEMERAL flag

        payload = InteractionResponsePayload(
            type=InteractionCallbackType.CHANNEL_MESSAGE_WITH_SOURCE,
            data=InteractionCallbackData(data),
        )

        await self._client._http.create_interaction_response(
            interaction_id=self.id,
            interaction_token=self.token,
            payload=payload,
        )

    async def respond_modal(self, modal: "Modal") -> None:
        """|coro| Send a modal in response to this interaction."""
        payload = InteractionResponsePayload(
            type=InteractionCallbackType.MODAL,
            data=modal.to_dict(),
        )
        await self._client._http.create_interaction_response(
            interaction_id=self.id,
            interaction_token=self.token,
            payload=payload,
        )

    async def edit(
        self,
        content: Optional[str] = None,
        *,
        embed: Optional[Embed] = None,
        embeds: Optional[List[Embed]] = None,
        components: Optional[List[ActionRow]] = None,
        attachments: Optional[List[Any]] = None,
        allowed_mentions: Optional[Dict[str, Any]] = None,
    ) -> None:
        """|coro|

        Edits the original response to this interaction.

        If the interaction is from a component, this will acknowledge the
        interaction and update the message in one operation.

        Parameters:
            content (Optional[str]): The new message content.
            embed (Optional[Embed]): A single embed to send. Ignored if
                ``embeds`` is provided.
            embeds (Optional[List[Embed]]): A list of embeds to send.
            components (Optional[List[ActionRow]]): Updated components for the
                message.
            attachments (Optional[List[Any]]): Attachments to include with the
                message.
            allowed_mentions (Optional[Dict[str, Any]]): Controls mentions in the
                message.
        """
        if embed and embeds:
            raise ValueError("Cannot provide both embed and embeds.")

        payload_data: Dict[str, Any] = {}
        if content is not None:
            payload_data["content"] = content
        if embed:
            payload_data["embeds"] = [embed.to_dict()]
        elif embeds is not None:
            payload_data["embeds"] = [e.to_dict() for e in embeds]
        if components is not None:
            payload_data["components"] = [c.to_dict() for c in components]
        if attachments is not None:
            payload_data["attachments"] = [
                a.to_dict() if hasattr(a, "to_dict") else a for a in attachments
            ]
        if allowed_mentions is not None:
            payload_data["allowed_mentions"] = allowed_mentions

        if self.type == InteractionType.MESSAGE_COMPONENT:
            # For component interactions, we send an UPDATE_MESSAGE response
            # to acknowledge the interaction and edit the message simultaneously.
            payload = InteractionResponsePayload(
                type=InteractionCallbackType.UPDATE_MESSAGE,
                data=InteractionCallbackData(payload_data),
            )
            await self._client._http.create_interaction_response(
                self.id, self.token, payload
            )
        else:
            # For other interaction types (like an initial slash command response),
            # we edit the original response via the webhook endpoint.
            await self._client._http.edit_original_interaction_response(
                application_id=self.application_id,
                interaction_token=self.token,
                payload=payload_data,
            )

    def __repr__(self) -> str:
        return f"<Interaction id='{self.id}' type={self.type!r}>"


class InteractionResponse:
    """Helper for sending responses for an :class:`Interaction`."""

    def __init__(self, interaction: "Interaction") -> None:
        self._interaction = interaction

    async def send_modal(self, modal: "Modal") -> None:
        """Sends a modal response."""
        payload = InteractionResponsePayload(
            type=InteractionCallbackType.MODAL,
            data=modal.to_dict(),
        )
        await self._interaction._client._http.create_interaction_response(
            self._interaction.id,
            self._interaction.token,
            payload,
        )


# Based on Interaction Response Object's Data Structure
class InteractionCallbackData:
    """Data for an interaction response."""

    def __init__(self, data: dict):
        self.tts: Optional[bool] = data.get("tts")
        self.content: Optional[str] = data.get("content")
        self.embeds: Optional[List[Embed]] = (
            [Embed(e) for e in data.get("embeds", [])] if data.get("embeds") else None
        )
        self.allowed_mentions: Optional[AllowedMentions] = (
            AllowedMentions(data["allowed_mentions"])
            if "allowed_mentions" in data
            else None
        )
        self.flags: Optional[int] = data.get("flags")  # MessageFlags enum could be used
        from .components import component_factory

        self.components: Optional[List[Component]] = (
            [component_factory(c) for c in data.get("components", [])]
            if data.get("components")
            else None
        )
        self.attachments: Optional[List[Attachment]] = (
            [Attachment(a) for a in data.get("attachments", [])]
            if data.get("attachments")
            else None
        )

    def to_dict(self) -> dict:
        # Helper to convert to dict for sending to Discord API
        payload = {}
        if self.tts is not None:
            payload["tts"] = self.tts
        if self.content is not None:
            payload["content"] = self.content
        if self.embeds is not None:
            payload["embeds"] = [e.to_dict() for e in self.embeds]
        if self.allowed_mentions is not None:
            payload["allowed_mentions"] = self.allowed_mentions.to_dict()
        if self.flags is not None:
            payload["flags"] = self.flags
        if self.components is not None:
            payload["components"] = [c.to_dict() for c in self.components]
        if self.attachments is not None:
            payload["attachments"] = [a.to_dict() for a in self.attachments]
        return payload

    def __repr__(self) -> str:
        return f"<InteractionCallbackData content='{self.content[:20] if self.content else None}'>"


# Based on Interaction Response Object Structure
class InteractionResponsePayload:
    """Payload for responding to an interaction."""

    def __init__(
        self,
        type: InteractionCallbackType,
        data: Optional[Union[InteractionCallbackData, Dict[str, Any]]] = None,
    ):
        self.type = type
        self.data = data

    def to_dict(self) -> Dict[str, Any]:
        payload: Dict[str, Any] = {"type": self.type.value}
        if self.data:
            if isinstance(self.data, dict):
                payload["data"] = self.data
            else:
                payload["data"] = self.data.to_dict()
        return payload

    def __repr__(self) -> str:
        return f"<InteractionResponsePayload type={self.type!r}>"

    def __getitem__(self, key: str) -> Any:
        return self.to_dict()[key]



import pytest

from disagreement.models import Embed


@pytest.mark.asyncio
async def test_edit_calls_http_with_payload(dummy_bot, interaction):
    await interaction.edit(content="updated")
    dummy_bot._http.edit_original_interaction_response.assert_called_once()
    kwargs = dummy_bot._http.edit_original_interaction_response.call_args.kwargs
    assert kwargs["application_id"] == dummy_bot.application_id
    assert kwargs["interaction_token"] == interaction.token
    assert kwargs["payload"] == {"content": "updated"}


@pytest.mark.asyncio
async def test_edit_embed_and_embeds_raises(dummy_bot, interaction):
    with pytest.raises(ValueError):
        await interaction.edit(embed=Embed(), embeds=[Embed()])



"""Audio source abstractions for the voice client."""

from __future__ import annotations

import asyncio
import contextlib
import io
import shlex
from typing import Optional, Union


class AudioSource:
    """Abstract base class for audio sources."""

    async def read(self) -> bytes:
        """Read the next chunk of PCM audio.

        Subclasses must implement this and return raw PCM data
        at 48kHz stereo (3840 byte chunks).
        """

        raise NotImplementedError

    async def close(self) -> None:
        """Cleanup the source when playback ends."""

        return None


class FFmpegAudioSource(AudioSource):
    """Decode audio using FFmpeg.

    Parameters
    ----------
    source:
        A filename, URL, or file-like object to read from.
    """

    def __init__(
        self,
        source: Union[str, io.BufferedIOBase],
        *,
        before_options: Optional[str] = None,
        options: Optional[str] = None,
        volume: float = 1.0,
    ):
        self.source = source
        self.before_options = before_options
        self.options = options
        self.volume = volume
        self.process: Optional[asyncio.subprocess.Process] = None
        self._feeder: Optional[asyncio.Task] = None

    async def _spawn(self) -> None:
        if isinstance(self.source, str):
            args = ["ffmpeg"]
            if self.before_options:
                args += shlex.split(self.before_options)
            args += [
                "-i",
                self.source,
                "-f",
                "s16le",
                "-ar",
                "48000",
                "-ac",
                "2",
                "pipe:1",
            ]
            if self.options:
                args += shlex.split(self.options)
            self.process = await asyncio.create_subprocess_exec(
                *args,
                stdout=asyncio.subprocess.PIPE,
                stderr=asyncio.subprocess.DEVNULL,
            )
        else:
            args = ["ffmpeg"]
            if self.before_options:
                args += shlex.split(self.before_options)
            args += [
                "-i",
                "pipe:0",
                "-f",
                "s16le",
                "-ar",
                "48000",
                "-ac",
                "2",
                "pipe:1",
            ]
            if self.options:
                args += shlex.split(self.options)
            self.process = await asyncio.create_subprocess_exec(
                *args,
                stdin=asyncio.subprocess.PIPE,
                stdout=asyncio.subprocess.PIPE,
                stderr=asyncio.subprocess.DEVNULL,
            )
            assert self.process.stdin is not None
            self._feeder = asyncio.create_task(self._feed())

    async def _feed(self) -> None:
        assert isinstance(self.source, io.BufferedIOBase)
        assert self.process is not None
        assert self.process.stdin is not None
        while True:
            data = await asyncio.to_thread(self.source.read, 4096)
            if not data:
                break
            self.process.stdin.write(data)
            await self.process.stdin.drain()
        self.process.stdin.close()

    async def read(self) -> bytes:
        if self.process is None:
            await self._spawn()
        assert self.process is not None
        assert self.process.stdout is not None
        data = await self.process.stdout.read(3840)
        if not data:
            await self.close()
        return data

    async def close(self) -> None:
        if self._feeder:
            self._feeder.cancel()
            with contextlib.suppress(asyncio.CancelledError):
                await self._feeder
        if self.process:
            await self.process.wait()
            self.process = None
        if isinstance(self.source, io.IOBase):
            with contextlib.suppress(Exception):
                self.source.close()


class AudioSink:
    """Abstract base class for audio sinks."""

    def write(self, user, data):
        """Write a chunk of PCM audio.

        Subclasses must implement this. The data is raw PCM at 48kHz
        stereo.
        """

        raise NotImplementedError

    def close(self) -> None:
        """Cleanup the sink when the voice client disconnects."""

        return None



from .view import View
from .item import Item
from .button import Button, button
from .select import Select, select
from .modal import Modal, TextInput, text_input

__all__ = [
    "View",
    "Item",
    "Button",
    "button",
    "Select",
    "select",
    "Modal",
    "TextInput",
    "text_input",
]



"""
Event dispatcher for handling Discord Gateway events.
"""

import asyncio
import inspect
from collections import defaultdict
from typing import (
    Callable,
    Coroutine,
    Any,
    Dict,
    List,
    Set,
    TYPE_CHECKING,
    Awaitable,
    Optional,
)

from .models import Message, User  # Assuming User might be part of other events
from .errors import DisagreementException

if TYPE_CHECKING:
    from .client import Client  # For type hinting to avoid circular imports
    from .interactions import Interaction

# Type alias for an event listener
EventListener = Callable[..., Awaitable[None]]


class EventDispatcher:
    """
    Manages registration and dispatching of event listeners.
    """

    def __init__(self, client_instance: "Client"):
        self._client: "Client" = client_instance
        self._listeners: Dict[str, List[EventListener]] = defaultdict(list)
        self._waiters: Dict[
            str, List[tuple[asyncio.Future, Optional[Callable[[Any], bool]]]]
        ] = defaultdict(list)
        self.on_dispatch_error: Optional[
            Callable[[str, Exception, EventListener], Awaitable[None]]
        ] = None
        # Pre-defined parsers for specific event types to convert raw data to models
        self._event_parsers: Dict[str, Callable[[Dict[str, Any]], Any]] = {
            "MESSAGE_CREATE": self._parse_message_create,
            "MESSAGE_UPDATE": self._parse_message_update,
            "MESSAGE_DELETE": self._parse_message_delete,
            "MESSAGE_REACTION_ADD": self._parse_message_reaction,
            "MESSAGE_REACTION_REMOVE": self._parse_message_reaction,
            "INTERACTION_CREATE": self._parse_interaction_create,
            "GUILD_CREATE": self._parse_guild_create,
            "CHANNEL_CREATE": self._parse_channel_create,
            "CHANNEL_UPDATE": self._parse_channel_update,
            "PRESENCE_UPDATE": self._parse_presence_update,
            "GUILD_MEMBER_ADD": self._parse_guild_member_add,
            "GUILD_MEMBER_REMOVE": self._parse_guild_member_remove,
            "GUILD_BAN_ADD": self._parse_guild_ban_add,
            "GUILD_BAN_REMOVE": self._parse_guild_ban_remove,
            "GUILD_ROLE_UPDATE": self._parse_guild_role_update,
            "TYPING_START": self._parse_typing_start,
            "VOICE_STATE_UPDATE": self._parse_voice_state_update,
        }

    def _parse_message_create(self, data: Dict[str, Any]) -> Message:
        """Parses raw MESSAGE_CREATE data into a Message object."""
        return self._client.parse_message(data)

    def _parse_message_update(self, data: Dict[str, Any]) -> Message:
        """Parses raw MESSAGE_UPDATE data into a Message object."""
        return self._client.parse_message(data)

    def _parse_message_delete(self, data: Dict[str, Any]) -> Dict[str, Any]:
        """Parses MESSAGE_DELETE and updates message cache."""
        message_id = data.get("id")
        if message_id:
            self._client._messages.invalidate(message_id)
        return data

    def _parse_message_reaction_raw(self, data: Dict[str, Any]) -> Dict[str, Any]:
        """Returns the raw reaction payload."""
        return data

    def _parse_interaction_create(self, data: Dict[str, Any]) -> "Interaction":
        """Parses raw INTERACTION_CREATE data into an Interaction object."""
        from .interactions import Interaction

        return Interaction(data=data, client_instance=self._client)

    def _parse_guild_create(self, data: Dict[str, Any]):
        """Parses raw GUILD_CREATE data into a Guild object."""

        return self._client.parse_guild(data)

    def _parse_channel_create(self, data: Dict[str, Any]):
        """Parses raw CHANNEL_CREATE data into a Channel object."""

        return self._client.parse_channel(data)

    def _parse_presence_update(self, data: Dict[str, Any]):
        """Parses raw PRESENCE_UPDATE data into a PresenceUpdate object."""

        from .models import PresenceUpdate

        return PresenceUpdate(data, client_instance=self._client)

    def _parse_typing_start(self, data: Dict[str, Any]):
        """Parses raw TYPING_START data into a TypingStart object."""

        from .models import TypingStart

        return TypingStart(data, client_instance=self._client)

    def _parse_voice_state_update(self, data: Dict[str, Any]):
        """Parses raw VOICE_STATE_UPDATE data into a VoiceStateUpdate object."""

        from .models import VoiceStateUpdate

        return VoiceStateUpdate(data, client_instance=self._client)

    def _parse_message_reaction(self, data: Dict[str, Any]):
        """Parses raw reaction data into a Reaction object."""

        from .models import Reaction

        return Reaction(data, client_instance=self._client)

    def _parse_guild_member_add(self, data: Dict[str, Any]):
        """Parses GUILD_MEMBER_ADD into a Member object."""

        guild_id = str(data.get("guild_id"))
        return self._client.parse_member(data, guild_id, just_joined=True)

    def _parse_guild_member_remove(self, data: Dict[str, Any]):
        """Parses GUILD_MEMBER_REMOVE into a GuildMemberRemove model."""

        from .models import GuildMemberRemove

        return GuildMemberRemove(data, client_instance=self._client)

    def _parse_guild_ban_add(self, data: Dict[str, Any]):
        """Parses GUILD_BAN_ADD into a GuildBanAdd model."""

        from .models import GuildBanAdd

        return GuildBanAdd(data, client_instance=self._client)

    def _parse_guild_ban_remove(self, data: Dict[str, Any]):
        """Parses GUILD_BAN_REMOVE into a GuildBanRemove model."""

        from .models import GuildBanRemove

        return GuildBanRemove(data, client_instance=self._client)

    def _parse_channel_update(self, data: Dict[str, Any]):
        """Parses CHANNEL_UPDATE into a Channel object."""

        return self._client.parse_channel(data)

    def _parse_guild_role_update(self, data: Dict[str, Any]):
        """Parses GUILD_ROLE_UPDATE into a GuildRoleUpdate model."""

        from .models import GuildRoleUpdate

        return GuildRoleUpdate(data, client_instance=self._client)

    # Potentially add _parse_user for events that directly provide a full user object
    # def _parse_user_update(self, data: Dict[str, Any]) -> User:
    #     return User(data=data)

    def register(self, event_name: str, coro: EventListener):
        """
        Registers a coroutine function to listen for a specific event.

        Args:
            event_name (str): The name of the event (e.g., 'MESSAGE_CREATE').
            coro (Callable): The coroutine function to call when the event occurs.
                             It should accept arguments appropriate for the event.

        Raises:
            TypeError: If the provided callback is not a coroutine function.
        """
        if not inspect.iscoroutinefunction(coro):
            raise TypeError(
                f"Event listener for '{event_name}' must be a coroutine function (async def)."
            )

        # Normalize event name, e.g., 'on_message' -> 'MESSAGE_CREATE'
        # For now, we assume event_name is already the Discord event type string.
        # If using decorators like @client.on_message, the decorator would handle this mapping.
        self._listeners[event_name.upper()].append(coro)

    def unregister(self, event_name: str, coro: EventListener):
        """
        Unregisters a coroutine function from an event.

        Args:
            event_name (str): The name of the event.
            coro (Callable): The coroutine function to unregister.
        """
        event_name_upper = event_name.upper()
        if event_name_upper in self._listeners:
            try:
                self._listeners[event_name_upper].remove(coro)
            except ValueError:
                pass

    def add_waiter(
        self,
        event_name: str,
        future: asyncio.Future,
        check: Optional[Callable[[Any], bool]] = None,
    ) -> None:
        self._waiters[event_name.upper()].append((future, check))

    def remove_waiter(self, event_name: str, future: asyncio.Future) -> None:
        waiters = self._waiters.get(event_name.upper())
        if not waiters:
            return
        self._waiters[event_name.upper()] = [
            (f, c) for f, c in waiters if f is not future
        ]
        if not self._waiters[event_name.upper()]:
            self._waiters.pop(event_name.upper(), None)

    def _resolve_waiters(self, event_name: str, data: Any) -> None:
        waiters = self._waiters.get(event_name)
        if not waiters:
            return
        to_remove: List[tuple[asyncio.Future, Optional[Callable[[Any], bool]]]] = []
        for future, check in waiters:
            if future.cancelled():
                to_remove.append((future, check))
                continue
            try:
                if check is None or check(data):
                    future.set_result(data)
                    to_remove.append((future, check))
            except Exception as exc:
                future.set_exception(exc)
                to_remove.append((future, check))
        for item in to_remove:
            if item in waiters:
                waiters.remove(item)
        if not waiters:
            self._waiters.pop(event_name, None)

    async def _dispatch_to_listeners(self, event_name: str, data: Any) -> None:
        listeners = self._listeners.get(event_name)
        if not listeners:
            return

        self._resolve_waiters(event_name, data)

        for listener in listeners:
            try:
                sig = inspect.signature(listener)
                num_params = len(sig.parameters)

                if num_params == 0:
                    await listener()
                elif num_params == 1:
                    await listener(data)
                else:
                    print(
                        f"Warning: Listener {listener.__name__} for {event_name} has an unhandled number of parameters ({num_params}). Skipping or attempting with one arg."
                    )
                    if num_params > 0:
                        await listener(data)

            except Exception as e:
                callback = self.on_dispatch_error
                if callback is not None:
                    try:
                        await callback(event_name, e, listener)
                    except Exception as hook_error:
                        print(f"Error in on_dispatch_error hook itself: {hook_error}")
                else:
                    print(
                        f"Error in event listener {listener.__name__} for {event_name}: {e}"
                    )
                    if hasattr(self._client, "on_error"):
                        try:
                            await self._client.on_error(event_name, e, listener)
                        except Exception as client_err_e:
                            print(f"Error in client.on_error itself: {client_err_e}")

    async def dispatch(self, event_name: str, raw_data: Dict[str, Any]):
        """Dispatch an event and its raw counterpart to all listeners."""

        event_name_upper = event_name.upper()
        raw_event_name = f"RAW_{event_name_upper}"

        await self._dispatch_to_listeners(raw_event_name, raw_data)

        parsed_data: Any = raw_data
        if event_name_upper in self._event_parsers:
            try:
                parser = self._event_parsers[event_name_upper]
                parsed_data = parser(raw_data)
            except Exception as e:
                print(f"Error parsing event data for {event_name_upper}: {e}")
                return

        await self._dispatch_to_listeners(event_name_upper, parsed_data)



"""Example showing how to read a channel's message history."""

import asyncio
import os
import sys

# Allow running example from repository root
if os.path.join(os.getcwd(), "examples") == os.path.dirname(os.path.abspath(__file__)):
    sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), "..")))

from disagreement.client import Client
from disagreement.models import TextChannel

try:
    from dotenv import load_dotenv
except ImportError:  # pragma: no cover - example helper
    load_dotenv = None
    print("python-dotenv is not installed. Environment variables will not be loaded")

if load_dotenv:
    load_dotenv()

BOT_TOKEN = os.environ.get("DISCORD_BOT_TOKEN", "")
CHANNEL_ID = os.environ.get("DISCORD_CHANNEL_ID", "")

client = Client(token=BOT_TOKEN)


async def main() -> None:
    channel = await client.fetch_channel(CHANNEL_ID)
    if isinstance(channel, TextChannel):
        async for message in channel.history(limit=10):
            print(message.content)


if __name__ == "__main__":
    asyncio.run(main())



"""
Disagreement
~~~~~~~~~~~~

A Python library for interacting with the Discord API.

:copyright: (c) 2025 Slipstream
:license: BSD 3-Clause License, see LICENSE for more details.
"""

__title__ = "disagreement"
__author__ = "Slipstream"
__license__ = "BSD 3-Clause License"
__copyright__ = "Copyright 2025 Slipstream"
__version__ = "0.4.2"

from .client import Client, AutoShardedClient
from .models import Message, User, Reaction, AuditLogEntry
from .voice_client import VoiceClient
from .audio import AudioSource, FFmpegAudioSource
from .typing import Typing
from .errors import (
    DisagreementException,
    HTTPException,
    GatewayException,
    AuthenticationError,
    Forbidden,
    NotFound,
)
from .color import Color
from .utils import utcnow, message_pager
from .enums import GatewayIntent, GatewayOpcode
from .error_handler import setup_global_error_handler
from .hybrid_context import HybridContext
from .ext import tasks
from .logging_config import setup_logging

import logging


__all__ = [
    "Client",
    "AutoShardedClient",
    "Message",
    "User",
    "Reaction",
    "AuditLogEntry",
    "VoiceClient",
    "AudioSource",
    "FFmpegAudioSource",
    "Typing",
    "DisagreementException",
    "HTTPException",
    "GatewayException",
    "AuthenticationError",
    "Forbidden",
    "NotFound",
    "Color",
    "utcnow",
    "message_pager",
    "GatewayIntent",
    "GatewayOpcode",
    "setup_global_error_handler",
    "HybridContext",
    "tasks",
    "setup_logging",
]


# Configure a default logger if none has been configured yet
if not logging.getLogger().hasHandlers():
    setup_logging(logging.INFO)



import pytest
from disagreement.event_dispatcher import EventDispatcher
from disagreement.models import Reaction


@pytest.mark.asyncio
async def test_reaction_payload():
    data = {
        "user_id": "1",
        "channel_id": "2",
        "message_id": "3",
        "emoji": {"name": "😀", "id": None},
    }
    reaction = Reaction(data)
    assert reaction.user_id == "1"
    assert reaction.emoji["name"] == "😀"


@pytest.mark.asyncio
async def test_dispatch_reaction_event(dummy_client):
    dispatcher = EventDispatcher(dummy_client)
    captured = []

    async def listener(payload: Reaction):
        captured.append(payload)

    # The event name is now MESSAGE_REACTION_ADD as per the original test setup.
    # If this were to fail, the next step would be to confirm the correct event name.
    dispatcher.register("MESSAGE_REACTION_ADD", listener)
    payload = {
        "user_id": "1",
        "channel_id": "2",
        "message_id": "3",
        "emoji": {"name": "👍", "id": None},
    }
    await dispatcher.dispatch("MESSAGE_REACTION_ADD", payload)
    assert len(captured) == 1
    assert isinstance(captured[0], Reaction)



"""Utility class for working with either command or app contexts."""

from __future__ import annotations

from typing import Any, Union

from .ext.commands.core import CommandContext
from .ext.app_commands.context import AppCommandContext


class HybridContext:
    """Wraps :class:`CommandContext` and :class:`AppCommandContext`.

    Provides a single :meth:`send` method that proxies to ``reply`` for
    prefix commands and to ``send`` for slash commands.
    """

    def __init__(self, ctx: Union[CommandContext, AppCommandContext]):
        self._ctx = ctx

    async def send(self, *args: Any, **kwargs: Any):
        if isinstance(self._ctx, AppCommandContext):
            return await self._ctx.send(*args, **kwargs)
        return await self._ctx.reply(*args, **kwargs)

    async def edit(self, *args: Any, **kwargs: Any):
        if hasattr(self._ctx, "edit"):
            return await self._ctx.edit(*args, **kwargs)
        raise AttributeError("Underlying context does not support editing.")

    def __getattr__(self, name: str) -> Any:
        return getattr(self._ctx, name)



import pytest

from disagreement.ui import Modal, TextInput


class MyModal(Modal):
    def __init__(self):
        super().__init__(title="T", custom_id="m")
        self.input = TextInput(label="L", custom_id="i")


@pytest.mark.asyncio
async def test_send_modal(dummy_bot, interaction):
    modal = MyModal()
    await interaction.response.send_modal(modal)
    dummy_bot._http.create_interaction_response.assert_called_once()



"""Example bot demonstrating gateway sharding."""

import asyncio
import os
import sys

# Ensure local package is importable when running from the examples directory
if os.path.join(os.getcwd(), "examples") == os.path.dirname(os.path.abspath(__file__)):
    sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), "..")))

import disagreement

try:
    from dotenv import load_dotenv
except ImportError:  # pragma: no cover - example helper
    load_dotenv = None
    print("python-dotenv is not installed. Environment variables will not be loaded")

if load_dotenv:
    load_dotenv()

TOKEN = os.environ.get("DISCORD_BOT_TOKEN")
if not TOKEN:
    raise RuntimeError("DISCORD_BOT_TOKEN environment variable not set")

client = disagreement.Client(token=TOKEN, shard_count=2)


@client.event
async def on_ready():
    if client.user:
        print(f"Shard bot ready as {client.user.username}#{client.user.discriminator}")
    else:
        print("Shard bot ready")


async def main():
    if not TOKEN:
        print("DISCORD_BOT_TOKEN environment variable not set")
        return
    await client.run()


if __name__ == "__main__":
    asyncio.run(main())



from __future__ import annotations

import asyncio
import logging
import inspect
from typing import (
    TYPE_CHECKING,
    Optional,
    List,
    Dict,
    Any,
    Union,
    Callable,
    Awaitable,
    Tuple,
    get_origin,
    get_args,
)

from .view import StringView
from .errors import (
    CommandError,
    CommandNotFound,
    BadArgument,
    MissingRequiredArgument,
    ArgumentParsingError,
    CheckFailure,
    CommandInvokeError,
)
from .converters import Greedy, run_converters, DEFAULT_CONVERTERS, Converter
from disagreement.typing import Typing

logger = logging.getLogger(__name__)

if TYPE_CHECKING:
    from .cog import Cog
    from disagreement.client import Client
    from disagreement.models import Message, User


class GroupMixin:
    def __init__(self, *args, **kwargs):
        super().__init__()
        self.commands: Dict[str, "Command"] = {}
        self.name: str = ""

    def command(
        self, **attrs: Any
    ) -> Callable[[Callable[..., Awaitable[None]]], "Command"]:
        def decorator(func: Callable[..., Awaitable[None]]) -> "Command":
            cmd = Command(func, **attrs)
            cmd.cog = getattr(self, "cog", None)
            self.add_command(cmd)
            return cmd

        return decorator

    def group(
        self, **attrs: Any
    ) -> Callable[[Callable[..., Awaitable[None]]], "Group"]:
        def decorator(func: Callable[..., Awaitable[None]]) -> "Group":
            cmd = Group(func, **attrs)
            cmd.cog = getattr(self, "cog", None)
            self.add_command(cmd)
            return cmd

        return decorator

    def add_command(self, command: "Command") -> None:
        if command.name in self.commands:
            raise ValueError(
                f"Command '{command.name}' is already registered in group '{self.name}'."
            )
        self.commands[command.name.lower()] = command
        for alias in command.aliases:
            if alias in self.commands:
                logger.warning(
                    f"Alias '{alias}' for command '{command.name}' in group '{self.name}' conflicts with an existing command or alias."
                )
            self.commands[alias.lower()] = command

    def get_command(self, name: str) -> Optional["Command"]:
        return self.commands.get(name.lower())


class Command(GroupMixin):
    """
    Represents a bot command.

    Attributes:
        name (str): The primary name of the command.
        callback (Callable[..., Awaitable[None]]): The coroutine function to execute.
        aliases (List[str]): Alternative names for the command.
        brief (Optional[str]): A short description for help commands.
        description (Optional[str]): A longer description for help commands.
        cog (Optional['Cog']): Reference to the Cog this command belongs to.
        params (Dict[str, inspect.Parameter]): Cached parameters of the callback.
    """

    def __init__(self, callback: Callable[..., Awaitable[None]], **attrs: Any):
        if not asyncio.iscoroutinefunction(callback):
            raise TypeError("Command callback must be a coroutine function.")

        super().__init__(**attrs)
        self.callback: Callable[..., Awaitable[None]] = callback
        self.name: str = attrs.get("name", callback.__name__)
        self.aliases: List[str] = attrs.get("aliases", [])
        self.brief: Optional[str] = attrs.get("brief")
        self.description: Optional[str] = attrs.get("description") or callback.__doc__
        self.cog: Optional["Cog"] = attrs.get("cog")
        self.invoke_without_command: bool = attrs.get("invoke_without_command", False)

        self.params = inspect.signature(callback).parameters
        self.checks: List[Callable[["CommandContext"], Awaitable[bool] | bool]] = []
        if hasattr(callback, "__command_checks__"):
            self.checks.extend(getattr(callback, "__command_checks__"))

        self.max_concurrency: Optional[Tuple[int, str]] = None
        if hasattr(callback, "__max_concurrency__"):
            self.max_concurrency = getattr(callback, "__max_concurrency__")

    def add_check(
        self, predicate: Callable[["CommandContext"], Awaitable[bool] | bool]
    ) -> None:
        self.checks.append(predicate)

    async def _run_checks(self, ctx: "CommandContext") -> None:
        """Runs all cog, local and global checks for the command."""
        from .errors import CheckFailure

        # Run cog-level check first
        if self.cog:
            cog_check = getattr(self.cog, "cog_check", None)
            if cog_check:
                try:
                    result = cog_check(ctx)
                    if inspect.isawaitable(result):
                        result = await result
                    if not result:
                        raise CheckFailure(
                            f"The cog-level check for command '{self.name}' failed."
                        )
                except CheckFailure:
                    raise
                except Exception as e:
                    raise CommandInvokeError(e) from e

        # Run local checks
        for predicate in self.checks:
            result = predicate(ctx)
            if inspect.isawaitable(result):
                result = await result
            if not result:
                raise CheckFailure(f"A local check for command '{self.name}' failed.")

        # Then run global checks from the handler
        if hasattr(ctx.bot, "command_handler"):
            for predicate in ctx.bot.command_handler._global_checks:
                result = predicate(ctx)
                if inspect.isawaitable(result):
                    result = await result
                if not result:
                    raise CheckFailure(
                        f"A global check failed for command '{self.name}'."
                    )

    async def invoke(self, ctx: "CommandContext", *args: Any, **kwargs: Any) -> None:
        await self._run_checks(ctx)

        before_invoke = None
        after_invoke = None

        if self.cog:
            before_invoke = getattr(self.cog, "cog_before_invoke", None)
            after_invoke = getattr(self.cog, "cog_after_invoke", None)

        if before_invoke:
            await before_invoke(ctx)

        try:
            if self.cog:
                await self.callback(self.cog, ctx, *args, **kwargs)
            else:
                await self.callback(ctx, *args, **kwargs)
        finally:
            if after_invoke:
                await after_invoke(ctx)


class Group(Command):
    """A command that can have subcommands."""

    def __init__(self, callback: Callable[..., Awaitable[None]], **attrs: Any):
        super().__init__(callback, **attrs)


PrefixCommand = Command  # Alias for clarity in hybrid commands


class CommandContext:
    """
    Represents the context in which a command is being invoked.
    """

    def __init__(
        self,
        *,
        message: "Message",
        bot: "Client",
        prefix: str,
        command: "Command",
        invoked_with: str,
        args: Optional[List[Any]] = None,
        kwargs: Optional[Dict[str, Any]] = None,
        cog: Optional["Cog"] = None,
    ):
        self.message: "Message" = message
        self.bot: "Client" = bot
        self.prefix: str = prefix
        self.command: "Command" = command
        self.invoked_with: str = invoked_with
        self.args: List[Any] = args or []
        self.kwargs: Dict[str, Any] = kwargs or {}
        self.cog: Optional["Cog"] = cog

        self.author: "User" = message.author

    @property
    def guild(self):
        """The guild this command was invoked in."""
        if self.message.guild_id and hasattr(self.bot, "get_guild"):
            return self.bot.get_guild(self.message.guild_id)
        return None

    async def reply(
        self,
        content: Optional[str] = None,
        *,
        mention_author: Optional[bool] = None,
        **kwargs: Any,
    ) -> "Message":
        """Replies to the invoking message.

        Parameters
        ----------
        content: str
            The content to send.
        mention_author: Optional[bool]
            Whether to mention the author in the reply. If ``None`` the
            client's :attr:`mention_replies` value is used.
        """

        allowed_mentions = kwargs.pop("allowed_mentions", None)
        if mention_author is None:
            mention_author = getattr(self.bot, "mention_replies", False)

        if allowed_mentions is None:
            allowed_mentions = {"replied_user": mention_author}
        else:
            allowed_mentions = dict(allowed_mentions)
            allowed_mentions.setdefault("replied_user", mention_author)

        return await self.bot.send_message(
            channel_id=self.message.channel_id,
            content=content,
            message_reference={
                "message_id": self.message.id,
                "channel_id": self.message.channel_id,
                "guild_id": self.message.guild_id,
            },
            allowed_mentions=allowed_mentions,
            **kwargs,
        )

    async def send(self, content: str, **kwargs: Any) -> "Message":
        return await self.bot.send_message(
            channel_id=self.message.channel_id, content=content, **kwargs
        )

    async def edit(
        self,
        message: Union[str, "Message"],
        *,
        content: Optional[str] = None,
        **kwargs: Any,
    ) -> "Message":
        """Edits a message previously sent by the bot."""

        message_id = message if isinstance(message, str) else message.id
        return await self.bot.edit_message(
            channel_id=self.message.channel_id,
            message_id=message_id,
            content=content,
            **kwargs,
        )

    def typing(self) -> "Typing":
        """Return a typing context manager for this context's channel."""

        return self.bot.typing(self.message.channel_id)


class CommandHandler:
    """
    Manages command registration, parsing, and dispatching.
    """

    def __init__(
        self,
        client: "Client",
        prefix: Union[
            str, List[str], Callable[["Client", "Message"], Union[str, List[str]]]
        ],
    ):
        self.client: "Client" = client
        self.prefix: Union[
            str, List[str], Callable[["Client", "Message"], Union[str, List[str]]]
        ] = prefix
        self.commands: Dict[str, Command] = {}
        self.cogs: Dict[str, "Cog"] = {}
        self._concurrency: Dict[str, Dict[str, int]] = {}
        self._global_checks: List[
            Callable[["CommandContext"], Awaitable[bool] | bool]
        ] = []

        from .help import HelpCommand

        self.add_command(HelpCommand(self))

    def add_check(
        self, predicate: Callable[["CommandContext"], Awaitable[bool] | bool]
    ) -> None:
        """Adds a global check to the command handler."""
        self._global_checks.append(predicate)

    def add_command(self, command: Command) -> None:
        if command.name in self.commands:
            raise ValueError(f"Command '{command.name}' is already registered.")

        self.commands[command.name.lower()] = command
        for alias in command.aliases:
            if alias in self.commands:
                logger.warning(
                    "Alias '%s' for command '%s' conflicts with an existing command or alias.",
                    alias,
                    command.name,
                )
            self.commands[alias.lower()] = command

        if isinstance(command, Group):
            for sub_cmd in command.commands.values():
                if sub_cmd.name in self.commands:
                    logger.warning(
                        "Subcommand '%s' of group '%s' conflicts with a top-level command.",
                        sub_cmd.name,
                        command.name,
                    )

    def remove_command(self, name: str) -> Optional[Command]:
        command = self.commands.pop(name.lower(), None)
        if command:
            for alias in command.aliases:
                self.commands.pop(alias.lower(), None)
        return command

    def get_command(self, name: str) -> Optional[Command]:
        return self.commands.get(name.lower())

    def add_cog(self, cog_to_add: "Cog") -> None:
        from .cog import Cog

        if not isinstance(cog_to_add, Cog):
            raise TypeError("Argument must be a subclass of Cog.")

        if cog_to_add.cog_name in self.cogs:
            raise ValueError(
                f"Cog with name '{cog_to_add.cog_name}' is already registered."
            )

        self.cogs[cog_to_add.cog_name] = cog_to_add

        for cmd in cog_to_add.get_commands():
            self.add_command(cmd)

        if hasattr(self.client, "_event_dispatcher"):
            for event_name, callback in cog_to_add.get_listeners():
                self.client._event_dispatcher.register(event_name.upper(), callback)
        else:
            logger.warning(
                "Client does not have '_event_dispatcher'. Listeners for cog '%s' not registered.",
                cog_to_add.cog_name,
            )

        if hasattr(cog_to_add, "cog_load") and inspect.iscoroutinefunction(
            cog_to_add.cog_load
        ):
            asyncio.create_task(cog_to_add.cog_load())

        logger.info("Cog '%s' added.", cog_to_add.cog_name)

    def remove_cog(self, cog_name: str) -> Optional["Cog"]:
        cog_to_remove = self.cogs.pop(cog_name, None)
        if cog_to_remove:
            for cmd in cog_to_remove.get_commands():
                self.remove_command(cmd.name)

            if hasattr(self.client, "_event_dispatcher"):
                for event_name, callback in cog_to_remove.get_listeners():
                    logger.debug(
                        "Listener '%s' for event '%s' from cog '%s' needs manual unregistration logic in EventDispatcher.",
                        callback.__name__,
                        event_name,
                        cog_name,
                    )

            if hasattr(cog_to_remove, "cog_unload") and inspect.iscoroutinefunction(
                cog_to_remove.cog_unload
            ):
                asyncio.create_task(cog_to_remove.cog_unload())

            cog_to_remove._eject()
            logger.info("Cog '%s' removed.", cog_name)
        return cog_to_remove

    def _acquire_concurrency(self, ctx: CommandContext) -> None:
        mc = getattr(ctx.command, "max_concurrency", None)
        if not mc:
            return
        limit, scope = mc
        if scope == "user":
            key = ctx.author.id
        elif scope == "guild":
            key = ctx.message.guild_id or ctx.author.id
        else:
            key = "global"
        buckets = self._concurrency.setdefault(ctx.command.name, {})
        current = buckets.get(key, 0)
        if current >= limit:
            from .errors import MaxConcurrencyReached

            raise MaxConcurrencyReached(limit)
        buckets[key] = current + 1

    def _release_concurrency(self, ctx: CommandContext) -> None:
        mc = getattr(ctx.command, "max_concurrency", None)
        if not mc:
            return
        _, scope = mc
        if scope == "user":
            key = ctx.author.id
        elif scope == "guild":
            key = ctx.message.guild_id or ctx.author.id
        else:
            key = "global"
        buckets = self._concurrency.get(ctx.command.name)
        if not buckets:
            return
        current = buckets.get(key, 0)
        if current <= 1:
            buckets.pop(key, None)
        else:
            buckets[key] = current - 1
        if not buckets:
            self._concurrency.pop(ctx.command.name, None)

    async def get_prefix(self, message: "Message") -> Union[str, List[str], None]:
        if callable(self.prefix):
            if inspect.iscoroutinefunction(self.prefix):
                return await self.prefix(self.client, message)
            else:
                return self.prefix(self.client, message)  # type: ignore
        return self.prefix

    async def _parse_arguments(
        self, command: Command, ctx: CommandContext, view: StringView
    ) -> Tuple[List[Any], Dict[str, Any]]:
        args_list = []
        kwargs_dict = {}
        params_to_parse = list(command.params.values())

        if params_to_parse and params_to_parse[0].name == "self" and command.cog:
            params_to_parse.pop(0)
        if params_to_parse and params_to_parse[0].name == "ctx":
            params_to_parse.pop(0)

        for param in params_to_parse:
            view.skip_whitespace()
            final_value_for_param: Any = inspect.Parameter.empty

            if param.kind == inspect.Parameter.VAR_POSITIONAL:
                while not view.eof:
                    view.skip_whitespace()
                    if view.eof:
                        break
                    word = view.get_word()
                    if word or not view.eof:
                        args_list.append(word)
                    elif view.eof:
                        break
                break

            arg_str_value: Optional[str] = (
                None  # Holds the raw string for current param
            )

            annotation = param.annotation
            if inspect.isclass(annotation) and issubclass(annotation, Greedy):
                greedy_values = []
                converter_type = annotation.converter
                while not view.eof:
                    view.skip_whitespace()
                    if view.eof:
                        break
                    start = view.index
                    if view.buffer[view.index] == '"':
                        arg_str_value = view.get_quoted_string()
                        if arg_str_value == "" and view.buffer[view.index] == '"':
                            raise BadArgument(
                                f"Unterminated quoted string for argument '{param.name}'."
                            )
                    else:
                        arg_str_value = view.get_word()
                    try:
                        converted = await run_converters(
                            ctx, converter_type, arg_str_value
                        )
                    except BadArgument:
                        view.index = start
                        break
                    greedy_values.append(converted)
                final_value_for_param = greedy_values
                arg_str_value = None
            elif view.eof:  # No more input string
                if param.default is not inspect.Parameter.empty:
                    final_value_for_param = param.default
                elif param.kind != inspect.Parameter.VAR_KEYWORD:
                    raise MissingRequiredArgument(param.name)
                else:  # VAR_KEYWORD at EOF is fine
                    break
            else:  # Input available
                is_last_pos_str_greedy = (
                    param == params_to_parse[-1]
                    and param.annotation is str
                    and param.kind == inspect.Parameter.POSITIONAL_OR_KEYWORD
                )

                if is_last_pos_str_greedy:
                    arg_str_value = view.read_rest().strip()
                    if (
                        not arg_str_value
                        and param.default is not inspect.Parameter.empty
                    ):
                        final_value_for_param = param.default
                    else:  # Includes empty string if that's what's left
                        final_value_for_param = arg_str_value
                else:  # Not greedy, or not string, or not last positional
                    if view.buffer[view.index] == '"':
                        arg_str_value = view.get_quoted_string()
                        if arg_str_value == "" and view.buffer[view.index] == '"':
                            raise BadArgument(
                                f"Unterminated quoted string for argument '{param.name}'."
                            )
                    else:
                        arg_str_value = view.get_word()

                # If final_value_for_param was not set by greedy logic, try conversion
                if final_value_for_param is inspect.Parameter.empty:
                    if arg_str_value is None:
                        if param.default is not inspect.Parameter.empty:
                            final_value_for_param = param.default
                        else:
                            raise MissingRequiredArgument(param.name)
                    else:  # We have an arg_str_value (could be empty string "" from quotes)
                        annotation = param.annotation
                        origin = get_origin(annotation)

                        if origin is Union:  # Handles Optional[T] and Union[T1, T2]
                            union_args = get_args(annotation)
                            is_optional = (
                                len(union_args) == 2 and type(None) in union_args
                            )

                            converted_for_union = False
                            last_err_union: Optional[BadArgument] = None
                            for t_arg in union_args:
                                if t_arg is type(None):
                                    continue
                                try:
                                    final_value_for_param = await run_converters(
                                        ctx, t_arg, arg_str_value
                                    )
                                    converted_for_union = True
                                    break
                                except BadArgument as e:
                                    last_err_union = e

                            if not converted_for_union:
                                if (
                                    is_optional and param.default is None
                                ):  # Special handling for Optional[T] if conversion failed
                                    # If arg_str_value was "" and type was Optional[str], StringConverter would return ""
                                    # If arg_str_value was "" and type was Optional[int], BadArgument would be raised.
                                    # This path is for when all actual types in Optional[T] fail conversion.
                                    # If default is None, we can assign None.
                                    final_value_for_param = None
                                elif last_err_union:
                                    raise last_err_union
                                else:
                                    raise BadArgument(
                                        f"Could not convert '{arg_str_value}' to any of {union_args} for param '{param.name}'."
                                    )
                        elif annotation is inspect.Parameter.empty or annotation is str:
                            final_value_for_param = arg_str_value
                        else:  # Standard type hint
                            final_value_for_param = await run_converters(
                                ctx, annotation, arg_str_value
                            )

            # Final check if value was resolved
            if final_value_for_param is inspect.Parameter.empty:
                if param.default is not inspect.Parameter.empty:
                    final_value_for_param = param.default
                elif param.kind != inspect.Parameter.VAR_KEYWORD:
                    # This state implies an issue if required and no default, and no input was parsed.
                    raise MissingRequiredArgument(
                        f"Parameter '{param.name}' could not be resolved."
                    )

            # Assign to args_list or kwargs_dict if a value was determined
            if final_value_for_param is not inspect.Parameter.empty:
                if (
                    param.kind == inspect.Parameter.POSITIONAL_OR_KEYWORD
                    or param.kind == inspect.Parameter.POSITIONAL_ONLY
                ):
                    args_list.append(final_value_for_param)
                elif param.kind == inspect.Parameter.KEYWORD_ONLY:
                    kwargs_dict[param.name] = final_value_for_param

        return args_list, kwargs_dict

    async def process_commands(self, message: "Message") -> None:
        if not message.content:
            return

        prefix_to_use = await self.get_prefix(message)
        if not prefix_to_use:
            return

        actual_prefix: Optional[str] = None
        if isinstance(prefix_to_use, list):
            for p in prefix_to_use:
                if message.content.startswith(p):
                    actual_prefix = p
                    break
            if not actual_prefix:
                return
        elif isinstance(prefix_to_use, str):
            if message.content.startswith(prefix_to_use):
                actual_prefix = prefix_to_use
            else:
                return
        else:
            return

        if actual_prefix is None:
            return

        content_without_prefix = message.content[len(actual_prefix) :]
        view = StringView(content_without_prefix)

        command_name = view.get_word()
        if not command_name:
            return

        command = self.get_command(command_name)
        if not command:
            return

        invoked_with = command_name
        original_command = command

        if isinstance(command, Group):
            view.skip_whitespace()
            potential_subcommand = view.get_word()
            if potential_subcommand:
                subcommand = command.get_command(potential_subcommand)
                if subcommand:
                    command = subcommand
                    invoked_with += f" {potential_subcommand}"
                elif command.invoke_without_command:
                    view.index -= len(potential_subcommand) + view.previous
                else:
                    raise CommandNotFound(
                        f"Subcommand '{potential_subcommand}' not found."
                    )

        ctx = CommandContext(
            message=message,
            bot=self.client,
            prefix=actual_prefix,
            command=command,
            invoked_with=invoked_with,
            cog=command.cog,
        )

        try:
            parsed_args, parsed_kwargs = await self._parse_arguments(command, ctx, view)
            ctx.args = parsed_args
            ctx.kwargs = parsed_kwargs
            self._acquire_concurrency(ctx)
            try:
                await command.invoke(ctx, *parsed_args, **parsed_kwargs)
            finally:
                self._release_concurrency(ctx)
        except CommandError as e:
            logger.error("Command error for '%s': %s", original_command.name, e)
            if hasattr(self.client, "on_command_error"):
                await self.client.on_command_error(ctx, e)
        except Exception as e:
            logger.error(
                "Unexpected error invoking command '%s': %s", original_command.name, e
            )
            exc = CommandInvokeError(e)
            if hasattr(self.client, "on_command_error"):
                await self.client.on_command_error(ctx, exc)
        else:
            if hasattr(self.client, "on_command_completion"):
                await self.client.on_command_completion(ctx)



import asyncio
import pytest
from unittest.mock import AsyncMock

from disagreement.client import Client


@pytest.mark.asyncio
async def test_client_async_context_closes(monkeypatch):
    client = Client(token="t")
    monkeypatch.setattr(client, "connect", AsyncMock())
    monkeypatch.setattr(client._http, "close", AsyncMock())

    async with client:
        client.connect.assert_awaited_once()

    client._http.close.assert_awaited_once()
    assert client.is_closed()


@pytest.mark.asyncio
async def test_client_async_context_closes_on_exception(monkeypatch):
    client = Client(token="t")
    monkeypatch.setattr(client, "connect", AsyncMock())
    monkeypatch.setattr(client._http, "close", AsyncMock())

    with pytest.raises(ValueError):
        async with client:
            raise ValueError("boom")

    client.connect.assert_awaited_once()
    client._http.close.assert_awaited_once()
    assert client.is_closed()



import pytest
from unittest.mock import AsyncMock

from disagreement.http import HTTPClient


@pytest.mark.asyncio
async def test_create_followup_message_calls_request():
    http = HTTPClient(token="t")
    http.request = AsyncMock()
    payload = {"content": "hello"}
    await http.create_followup_message("app_id", "token", payload)
    http.request.assert_called_once_with(
        "POST",
        f"/webhooks/app_id/token",
        payload=payload,
        use_auth_header=False,
    )


@pytest.mark.asyncio
async def test_edit_followup_message_calls_request():
    http = HTTPClient(token="t")
    http.request = AsyncMock()
    payload = {"content": "new content"}
    await http.edit_followup_message("app_id", "token", "123", payload)
    http.request.assert_called_once_with(
        "PATCH",
        f"/webhooks/app_id/token/messages/123",
        payload=payload,
        use_auth_header=False,
    )


@pytest.mark.asyncio
async def test_delete_followup_message_calls_request():
    http = HTTPClient(token="t")
    http.request = AsyncMock()
    await http.delete_followup_message("app_id", "token", "456")
    http.request.assert_called_once_with(
        "DELETE",
        f"/webhooks/app_id/token/messages/456",
        use_auth_header=False,
    )


@pytest.mark.asyncio
async def test_create_webhook_returns_model_and_calls_request():
    http = HTTPClient(token="t")
    http.request = AsyncMock(return_value={"id": "1"})
    payload = {"name": "wh"}
    webhook = await http.create_webhook("123", payload)
    http.request.assert_called_once_with(
        "POST",
        "/channels/123/webhooks",
        payload=payload,
    )
    from disagreement.models import Webhook

    assert isinstance(webhook, Webhook)


@pytest.mark.asyncio
async def test_edit_webhook_returns_model_and_calls_request():
    http = HTTPClient(token="t")
    http.request = AsyncMock(return_value={"id": "1"})
    payload = {"name": "rename"}
    webhook = await http.edit_webhook("1", payload)
    http.request.assert_called_once_with(
        "PATCH",
        "/webhooks/1",
        payload=payload,
    )
    from disagreement.models import Webhook

    assert isinstance(webhook, Webhook)


@pytest.mark.asyncio
async def test_delete_webhook_calls_request():
    http = HTTPClient(token="t")
    http.request = AsyncMock()
    await http.delete_webhook("1")
    http.request.assert_called_once_with(
        "DELETE",
        "/webhooks/1",
    )


@pytest.mark.asyncio
async def test_client_create_webhook_returns_model():
    from types import SimpleNamespace
    from disagreement.client import Client
    from disagreement.models import Webhook

    http = SimpleNamespace(create_webhook=AsyncMock(return_value={"id": "1"}))
    client = Client(token="test")
    client._http = http

    webhook = await client.create_webhook("123", {"name": "wh"})

    http.create_webhook.assert_awaited_once_with("123", {"name": "wh"})
    assert isinstance(webhook, Webhook)
    assert client._webhooks.get("1") is webhook


@pytest.mark.asyncio
async def test_client_edit_webhook_returns_model():
    from types import SimpleNamespace
    from disagreement.client import Client
    from disagreement.models import Webhook

    http = SimpleNamespace(edit_webhook=AsyncMock(return_value={"id": "1"}))
    client = Client(token="test")
    client._http = http

    webhook = await client.edit_webhook("1", {"name": "rename"})

    http.edit_webhook.assert_awaited_once_with("1", {"name": "rename"})
    assert isinstance(webhook, Webhook)
    assert client._webhooks.get("1") is webhook


@pytest.mark.asyncio
async def test_client_delete_webhook_calls_http():
    from types import SimpleNamespace
    from disagreement.client import Client

    http = SimpleNamespace(delete_webhook=AsyncMock())
    client = Client(token="test")
    client._http = http

    await client.delete_webhook("1")

    http.delete_webhook.assert_awaited_once_with("1")


def test_webhook_from_url_parses_id_and_token():
    from disagreement.models import Webhook

    url = "https://discord.com/api/webhooks/123/token"
    webhook = Webhook.from_url(url)

    assert webhook.id == "123"
    assert webhook.token == "token"
    assert webhook.url == url


@pytest.mark.asyncio
async def test_execute_webhook_calls_request():
    http = HTTPClient(token="t")
    http.request = AsyncMock(return_value={"id": "1"})
    await http.execute_webhook("1", "tok", content="hi")
    http.request.assert_called_once_with(
        "POST",
        "/webhooks/1/tok",
        payload={"content": "hi"},
        use_auth_header=False,
    )


@pytest.mark.asyncio
async def test_webhook_send_uses_http():
    from types import SimpleNamespace
    from disagreement.client import Client
    from disagreement.models import Webhook, Message

    http = SimpleNamespace(
        execute_webhook=AsyncMock(
            return_value={
                "id": "2",
                "channel_id": "c",
                "author": {"id": "1", "username": "u", "discriminator": "0001"},
                "content": "hi",
                "timestamp": "t",
            }
        )
    )
    client = Client(token="test")
    client._http = http

    webhook = Webhook({"id": "1", "token": "tok"}, client_instance=client)

    msg = await webhook.send(content="hi")

    http.execute_webhook.assert_awaited_once()
    assert isinstance(msg, Message)



"""Utility helpers for working with Discord permission bitmasks."""

from __future__ import annotations

from enum import IntFlag
from typing import Iterable, List


class Permissions(IntFlag):
    """Discord guild and channel permissions."""

    CREATE_INSTANT_INVITE = 1 << 0
    KICK_MEMBERS = 1 << 1
    BAN_MEMBERS = 1 << 2
    ADMINISTRATOR = 1 << 3
    MANAGE_CHANNELS = 1 << 4
    MANAGE_GUILD = 1 << 5
    ADD_REACTIONS = 1 << 6
    VIEW_AUDIT_LOG = 1 << 7
    PRIORITY_SPEAKER = 1 << 8
    STREAM = 1 << 9
    VIEW_CHANNEL = 1 << 10
    SEND_MESSAGES = 1 << 11
    SEND_TTS_MESSAGES = 1 << 12
    MANAGE_MESSAGES = 1 << 13
    EMBED_LINKS = 1 << 14
    ATTACH_FILES = 1 << 15
    READ_MESSAGE_HISTORY = 1 << 16
    MENTION_EVERYONE = 1 << 17
    USE_EXTERNAL_EMOJIS = 1 << 18
    VIEW_GUILD_INSIGHTS = 1 << 19
    CONNECT = 1 << 20
    SPEAK = 1 << 21
    MUTE_MEMBERS = 1 << 22
    DEAFEN_MEMBERS = 1 << 23
    MOVE_MEMBERS = 1 << 24
    USE_VAD = 1 << 25
    CHANGE_NICKNAME = 1 << 26
    MANAGE_NICKNAMES = 1 << 27
    MANAGE_ROLES = 1 << 28
    MANAGE_WEBHOOKS = 1 << 29
    MANAGE_GUILD_EXPRESSIONS = 1 << 30
    USE_APPLICATION_COMMANDS = 1 << 31
    REQUEST_TO_SPEAK = 1 << 32
    MANAGE_EVENTS = 1 << 33
    MANAGE_THREADS = 1 << 34
    CREATE_PUBLIC_THREADS = 1 << 35
    CREATE_PRIVATE_THREADS = 1 << 36
    USE_EXTERNAL_STICKERS = 1 << 37
    SEND_MESSAGES_IN_THREADS = 1 << 38
    USE_EMBEDDED_ACTIVITIES = 1 << 39
    MODERATE_MEMBERS = 1 << 40
    VIEW_CREATOR_MONETIZATION_ANALYTICS = 1 << 41
    USE_SOUNDBOARD = 1 << 42
    CREATE_GUILD_EXPRESSIONS = 1 << 43
    CREATE_EVENTS = 1 << 44
    USE_EXTERNAL_SOUNDS = 1 << 45
    SEND_VOICE_MESSAGES = 1 << 46


def permissions_value(*perms: Permissions | int | Iterable[Permissions | int]) -> int:
    """Return a combined integer value for multiple permissions."""

    value = 0
    for perm in perms:
        if isinstance(perm, Iterable) and not isinstance(perm, (Permissions, int)):
            value |= permissions_value(*perm)
        else:
            value |= int(perm)
    return value


def has_permissions(
    current: int | str | Permissions,
    *perms: Permissions | int | Iterable[Permissions | int],
) -> bool:
    """Return ``True`` if ``current`` includes all ``perms``."""

    current_val = int(current)
    needed = permissions_value(*perms)
    return (current_val & needed) == needed


def missing_permissions(
    current: int | str | Permissions,
    *perms: Permissions | int | Iterable[Permissions | int],
) -> List[Permissions]:
    """Return the subset of ``perms`` not present in ``current``."""

    current_val = int(current)
    missing: List[Permissions] = []
    for perm in perms:
        if isinstance(perm, Iterable) and not isinstance(perm, (Permissions, int)):
            missing.extend(missing_permissions(current_val, *perm))
        else:
            perm_val = int(perm)
            if not current_val & perm_val:
                missing.append(Permissions(perm_val))
    return missing



import json
from typing import Dict, Optional

_translations: Dict[str, Dict[str, str]] = {}


def set_translations(locale: str, mapping: Dict[str, str]) -> None:
    """Set translations for a locale."""
    _translations[locale] = mapping


def load_translations(locale: str, file_path: str) -> None:
    """Load translations for *locale* from a JSON file."""
    with open(file_path, "r", encoding="utf-8") as handle:
        _translations[locale] = json.load(handle)


def translate(key: str, locale: str, *, default: Optional[str] = None) -> str:
    """Return the translated string for *key* in *locale*."""
    return _translations.get(locale, {}).get(
        key, default if default is not None else key
    )



"""
disagreement.ext.commands - A command framework extension for the Disagreement library.
"""

from .cog import Cog
from .core import (
    Command,
    CommandContext,
    CommandHandler,
)  # CommandHandler might be internal
from .decorators import (
    command,
    listener,
    check,
    check_any,
    cooldown,
    max_concurrency,
    requires_permissions,
    has_role,
    has_any_role,
)
from .errors import (
    CommandError,
    CommandNotFound,
    BadArgument,
    MissingRequiredArgument,
    ArgumentParsingError,
    CheckFailure,
    CheckAnyFailure,
    CommandOnCooldown,
    CommandInvokeError,
    MaxConcurrencyReached,
)

__all__ = [
    # Cog
    "Cog",
    # Core
    "Command",
    "CommandContext",
    # "CommandHandler", # Usually not part of public API for direct use by bot devs
    # Decorators
    "command",
    "listener",
    "check",
    "check_any",
    "cooldown",
    "max_concurrency",
    "requires_permissions",
    "has_role",
    "has_any_role",
    # Errors
    "CommandError",
    "CommandNotFound",
    "BadArgument",
    "MissingRequiredArgument",
    "ArgumentParsingError",
    "CheckFailure",
    "CheckAnyFailure",
    "CommandOnCooldown",
    "CommandInvokeError",
    "MaxConcurrencyReached",
]



import pytest  # pylint: disable=import-error

from disagreement.i18n import set_translations, translate
from disagreement.ext.app_commands.commands import SlashCommand


async def dummy(ctx):
    pass


def test_translate_lookup():
    set_translations("xx", {"hello": "bonjour"})
    assert translate("hello", "xx") == "bonjour"
    assert translate("missing", "xx") == "missing"


def test_appcommand_uses_locale():
    set_translations("xx", {"cmd": "c", "desc": "d"})
    cmd = SlashCommand(dummy, name="cmd", description="desc", locale="xx")
    assert cmd.name == "c"
    assert cmd.description == "d"



from typing import List, Optional

from .core import Command, CommandContext, CommandHandler


class HelpCommand(Command):
    """Built-in command that displays help information for other commands."""

    def __init__(self, handler: CommandHandler) -> None:
        self.handler = handler

        async def callback(ctx: CommandContext, command: Optional[str] = None) -> None:
            if command:
                cmd = handler.get_command(command)
                if not cmd or cmd.name.lower() != command.lower():
                    await ctx.send(f"Command '{command}' not found.")
                    return
                description = cmd.description or cmd.brief or "No description provided."
                await ctx.send(f"**{ctx.prefix}{cmd.name}**\n{description}")
            else:
                lines: List[str] = []
                for registered in dict.fromkeys(handler.commands.values()):
                    brief = registered.brief or registered.description or ""
                    lines.append(f"{ctx.prefix}{registered.name} - {brief}".strip())
                if lines:
                    await ctx.send("\n".join(lines))
                else:
                    await ctx.send("No commands available.")

        super().__init__(
            callback,
            name="help",
            brief="Show command help.",
            description="Displays help for commands.",
        )



import pytest

from disagreement.ext.app_commands.handler import AppCommandHandler
from disagreement.ext.app_commands.decorators import user_command, message_command
from disagreement.enums import ApplicationCommandType, InteractionType
from disagreement.interactions import Interaction
from disagreement.models import User, Message


@pytest.mark.asyncio
async def test_user_context_menu_invokes(dummy_bot):
    handler = AppCommandHandler(dummy_bot)
    captured = {}

    @user_command(name="Info")
    async def info(ctx, user: User):
        captured["user"] = user

    handler.add_command(info)

    data = {
        "id": "cmd",
        "name": "Info",
        "type": ApplicationCommandType.USER.value,
        "target_id": "42",
        "resolved": {
            "users": {"42": {"id": "42", "username": "Bob", "discriminator": "0001"}}
        },
    }
    payload = {
        "id": "1",
        "application_id": dummy_bot.application_id,
        "type": InteractionType.APPLICATION_COMMAND.value,
        "token": "tok",
        "version": 1,
        "data": data,
    }
    interaction = Interaction(payload, client_instance=dummy_bot)
    await handler.process_interaction(interaction)
    assert isinstance(captured.get("user"), User)
    assert captured["user"].id == "42"


@pytest.mark.asyncio
async def test_message_context_menu_invokes(dummy_bot):
    handler = AppCommandHandler(dummy_bot)
    captured = {}

    @message_command(name="Quote")
    async def quote(ctx, message: Message):
        captured["msg"] = message

    handler.add_command(quote)

    msg_data = {
        "id": "99",
        "channel_id": "c",
        "author": {"id": "2", "username": "Ann", "discriminator": "0001"},
        "content": "Hello",
        "timestamp": "t",
    }
    data = {
        "id": "cmd",
        "name": "Quote",
        "type": ApplicationCommandType.MESSAGE.value,
        "target_id": "99",
        "resolved": {"messages": {"99": msg_data}},
    }
    payload = {
        "id": "1",
        "application_id": dummy_bot.application_id,
        "type": InteractionType.APPLICATION_COMMAND.value,
        "token": "tok",
        "version": 1,
        "data": data,
    }
    interaction = Interaction(payload, client_instance=dummy_bot)
    await handler.process_interaction(interaction)
    assert isinstance(captured.get("msg"), Message)
    assert captured["msg"].id == "99"



from __future__ import annotations

import asyncio
from typing import Any, Callable, Coroutine, List, Optional, TYPE_CHECKING

from .item import Item
from ..enums import ComponentType
from ..models import SelectOption

if TYPE_CHECKING:
    from ..interactions import Interaction


class Select(Item):
    """Represents a select menu component in a View.

    Args:
        custom_id (str): The developer-defined identifier for the select menu.
        options (List[SelectOption]): The choices in the select menu.
        placeholder (Optional[str]): The placeholder text that is shown if nothing is selected.
        min_values (int): The minimum number of items that must be chosen.
        max_values (int): The maximum number of items that can be chosen.
        disabled (bool): Whether the select menu is disabled.
        row (Optional[int]): The row the select menu should be placed in, from 0 to 4.
    """

    def __init__(
        self,
        *,
        custom_id: str,
        options: List[SelectOption],
        placeholder: Optional[str] = None,
        min_values: int = 1,
        max_values: int = 1,
        disabled: bool = False,
        row: Optional[int] = None,
    ):
        super().__init__(type=ComponentType.STRING_SELECT)
        self.custom_id = custom_id
        self.options = options
        self.placeholder = placeholder
        self.min_values = min_values
        self.max_values = max_values
        self.disabled = disabled
        self._row = row

    def to_dict(self) -> dict[str, Any]:
        """Converts the select menu to a dictionary that can be sent to Discord."""
        payload = {
            "type": ComponentType.STRING_SELECT.value,
            "custom_id": self.custom_id,
            "options": [option.to_dict() for option in self.options],
            "disabled": self.disabled,
        }
        if self.placeholder:
            payload["placeholder"] = self.placeholder
        if self.min_values is not None:
            payload["min_values"] = self.min_values
        if self.max_values is not None:
            payload["max_values"] = self.max_values
        return payload


def select(
    *,
    custom_id: str,
    options: List[SelectOption],
    placeholder: Optional[str] = None,
    min_values: int = 1,
    max_values: int = 1,
    disabled: bool = False,
    row: Optional[int] = None,
) -> Callable[[Callable[..., Coroutine[Any, Any, Any]]], Select]:
    """A decorator to create a select menu in a View."""

    def decorator(func: Callable[..., Coroutine[Any, Any, Any]]) -> Select:
        if not asyncio.iscoroutinefunction(func):
            raise TypeError("Select callback must be a coroutine function.")

        item = Select(
            custom_id=custom_id,
            options=options,
            placeholder=placeholder,
            min_values=min_values,
            max_values=max_values,
            disabled=disabled,
            row=row,
        )
        item.callback = func
        return item

    return decorator
